# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""SSH into the toil appliance container running on the leader of the cluster."""
import argparse
import logging
import sys
from typing import List

from toil.common import parser_with_common_options
from toil.provisioners import cluster_factory
from toil.statsAndLogging import set_logging_from_options

logger = logging.getLogger(__name__)


def main() -> None:
    parser = parser_with_common_options(provisioner_options=True, jobstore_option=False)
    parser.add_argument("--insecure", action='store_true',
                        help="Temporarily disable strict host key checking.")
    parser.add_argument("--sshOption", dest='sshOptions', default=[], action='append',
                        help="Pass an additional option to the SSH command.")
    parser.add_argument("--grafana_port", dest='grafana_port', default=3000,
                        help="Assign a local port to be used for the Grafana dashboard.")
    parser.add_argument('args', nargs=argparse.REMAINDER)
    options = parser.parse_args()
    set_logging_from_options(options)

    # Since we collect all the remaining arguments at the end for a command to
    # run, it's easy to lose options.
    if len(options.args) > 0 and options.args[0].startswith('-'):
        logger.warning('Argument \'%s\' interpreted as a command to run '
                       'despite looking like an option.', options.args[0])

    cluster = cluster_factory(provisioner=options.provisioner,
                              clusterName=options.clusterName,
                              zone=options.zone)
    command = options.args if options.args else ['bash']
    sshOptions: List[str] = options.sshOptions

    # Forward ports:
    # 3000 for Grafana dashboard
    # 9090 for Prometheus dashboard
    sshOptions.extend(['-L', f'{options.grafana_port}:localhost:3000',
                       '-L', '9090:localhost:9090'])

    cluster.getLeader().sshAppliance(*command, strict=not options.insecure, tty=sys.stdin.isatty(),
                                     sshOptions=sshOptions)



# Copyright (C) 2020-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from collections import OrderedDict
from typing import Union

from wdlparse.v1.WdlV1Lexer import FileStream, WdlV1Lexer
from wdlparse.v1.WdlV1Parser import CommonTokenStream, WdlV1Parser

from toil.wdl.wdl_analysis import AnalyzeWDL

logger = logging.getLogger(__name__)


def is_context(ctx, classname: Union[str, tuple]) -> bool:
    """
    Returns whether an ANTLR4 context object is of the precise type `classname`.

    :param ctx: An ANTLR4 context object.
    :param classname: The class name(s) as a string or a tuple of strings. If a
                      tuple is provided, this returns True if the context object
                      matches one of the class names.
    """
    # we check for `ctx.__class__.__name__` so that it's portable across multiple similar auto-generated parsers.
    if isinstance(classname, str):
        return ctx.__class__.__name__ == classname
    return ctx.__class__.__name__ in classname


class AnalyzeV1WDL(AnalyzeWDL):
    """
    AnalyzeWDL implementation for the 1.0 version using ANTLR4.

    See: https://github.com/openwdl/wdl/blob/main/versions/1.0/SPEC.md
         https://github.com/openwdl/wdl/blob/main/versions/1.0/parsers/antlr4/WdlV1Parser.g4
    """

    @property
    def version(self) -> str:
        return '1.0'

    def analyze(self):
        """
        Analyzes the WDL file passed into the constructor and generates the two
        intermediate data structures: `self.workflows_dictionary` and
        `self.tasks_dictionary`.
        """
        lexer = WdlV1Lexer(FileStream(self.wdl_file))
        parser = WdlV1Parser(input=CommonTokenStream(lexer))
        tree = parser.document()
        self.visit_document(tree)

    def visit_document(self, ctx):
        """
        Root of tree. Contains `version` followed by an optional workflow and
        any number of `document_element`s.
        """
        wf = ctx.workflow()
        if wf:
            self.visit_workflow(wf)

        for element in ctx.document_element():
            self.visit_document_element(element)

    def visit_document_element(self, ctx):
        """
        Contains one of the following: 'import_doc', 'struct', or 'task'.
        """
        element = ctx.children[0]
        # task
        if is_context(element, 'TaskContext'):
            return self.visit_task(element)
        # struct
        elif is_context(element, 'StructContext'):
            # TODO: add support for structs.
            raise NotImplementedError('Struct is not supported.')
        # import_doc
        elif is_context(element, 'Import_docContext'):
            # TODO: add support for imports.
            raise NotImplementedError('Import is not supported.')
        else:
            raise RuntimeError(f'Unrecognized document element in visitDocument(): {type(element)}')

    # Workflow section

    def visit_workflow(self, ctx):
        """
        Contains an 'identifier' and an array of `workflow_element`s.
        """
        identifier = ctx.Identifier().getText()
        wf = self.workflows_dictionary.setdefault(identifier, OrderedDict())

        for element in ctx.workflow_element():
            section = element.children[0]
            # input
            if is_context(section, 'Workflow_inputContext'):
                # loop through all inputs and add to the workflow dictionary.
                # treating this the same as workflow declarations for now
                for wf_input in self.visit_workflow_input(section):
                    wf[f'declaration{self.declaration_number}'] = wf_input
                    self.declaration_number += 1
            # output
            elif is_context(section, 'Workflow_outputContext'):
                # TODO: add support for workflow level outputs in wdl_synthesis
                wf['wf_outputs'] = self.visit_workflow_output(section)
            # inner_element
            # i.e.: non-input declarations, scatters, calls, and conditionals
            elif is_context(section, 'Inner_workflow_elementContext'):
                wf_key, contents = self.visit_inner_workflow_element(section)
                wf[wf_key] = contents
            # parameter_meta and meta
            elif is_context(section, ('Parameter_meta_elementContext', 'Meta_elementContext')):
                # ignore additional metadata information for now.
                pass
            else:
                raise RuntimeError(f'Unrecognized workflow element in visitWorkflow(): {type(section)}')

    def visit_workflow_input(self, ctx):
        """
        Contains an array of 'any_decls', which can be unbound or bound declarations.
        Example:
            input {
              String in_str = "twenty"
              Int in_int
            }

        Returns a list of tuple=(name, type, expr).
        """
        return [self.visit_any_decls(decl) for decl in ctx.any_decls()]

    def visit_workflow_output(self, ctx):
        """
        Contains an array of 'bound_decls' (unbound_decls not allowed).
        Example:
            output {
              String out_str = "output"
            }

        Returns a list of tuple=(name, type, expr).
        """
        return [self.visit_bound_decls(decl) for decl in ctx.bound_decls()]

    def visit_inner_workflow_element(self, ctx):
        """
        Returns a tuple=(unique_key, dict), where dict contains the contents of
        the given inner workflow element.
        """
        element = ctx.children[0]

        # bound_decls
        # i.e.: declarations declared outside of input section
        if is_context(element, 'Bound_declsContext'):
            key = f'declaration{self.declaration_number}'
            self.declaration_number += 1
            return key, self.visit_bound_decls(element)
        # call
        elif is_context(element, 'CallContext'):
            key = f'call{self.call_number}'
            self.call_number += 1
            return key, self.visit_call(element)
        # scatter
        elif is_context(element, 'ScatterContext'):
            key = f'scatter{self.scatter_number}'
            self.scatter_number += 1
            return key, self.visit_scatter(element)
        # conditional
        elif is_context(element, 'ConditionalContext'):
            key = f'if{self.if_number}'
            self.if_number += 1
            return key, self.visit_conditional(element)
        else:
            raise RuntimeError(f'Unrecognized workflow element in visitInner_workflow_element(): {type(element)}')

    def visit_call(self, ctx):
        """
        Pattern: CALL call_name call_alias? call_body?
        Example WDL syntax: call task_1 {input: arr=arr}

        Returns a dict={task, alias, io}.
        """
        name = '.'.join(identifier.getText() for identifier in ctx.call_name().Identifier())
        alias = ctx.call_alias().Identifier().getText() if ctx.call_alias() else name

        body = OrderedDict()
        # make sure that '{}' and '{input: ...}' are provided
        if ctx.call_body() and ctx.call_body().call_inputs():
            for input_ in ctx.call_body().call_inputs().call_input():
                body[input_.Identifier().getText()] = self.visit_expr(input_.expr())

        return {
            'task': name,
            'alias': alias,
            'io': body
        }

    def visit_scatter(self, ctx):
        """
        Pattern: SCATTER LPAREN Identifier In expr RPAREN LBRACE inner_workflow_element* RBRACE
        Example WDL syntax: scatter ( i in items) { ... }

        Returns a dict={item, collection, body}.
        """
        item = ctx.Identifier().getText()
        expr = self.visit_expr(ctx.expr())
        body = OrderedDict()
        for element in ctx.inner_workflow_element():
            body_key, contents = self.visit_inner_workflow_element(element)
            body[body_key] = contents
        return {
            'item': item,
            'collection': expr,
            'body': body
        }

    def visit_conditional(self, ctx):
        """
        Pattern: IF LPAREN expr RPAREN LBRACE inner_workflow_element* RBRACE
        Example WDL syntax: if (condition) { ... }

        Returns a dict={expression, body}.
        """
        # see https://github.com/openwdl/wdl/blob/main/versions/1.0/SPEC.md#conditionals
        expr = self.visit_expr(ctx.expr())

        body = OrderedDict()
        for element in ctx.inner_workflow_element():
            body_key, contents = self.visit_inner_workflow_element(element)
            body[body_key] = contents

        return {
            'expression': expr,
            'body': body
        }

    # Task section

    def visit_task(self, ctx):
        """
        Root of a task definition. Contains an `identifier` and an array of
        `task_element`s.
        """
        identifier = ctx.Identifier().getText()
        task = self.tasks_dictionary.setdefault(identifier, OrderedDict())
        # print(f'Visiting task: {identifier}')

        for element in ctx.task_element():
            section = element.children[0]
            # input
            if is_context(section, 'Task_inputContext'):
                task.setdefault('inputs', []).extend(self.visit_task_input(section))
            # output
            elif is_context(section, 'Task_outputContext'):
                task['outputs'] = self.visit_task_output(section)
            # command
            elif is_context(section, 'Task_commandContext'):
                task['raw_commandline'] = self.visit_task_command(section)
            # runtime
            elif is_context(section, 'Task_runtimeContext'):
                task['runtime'] = self.visit_task_runtime(section)
            # bound_decls
            elif is_context(section, 'Bound_declsContext'):
                # treating this the same as inputs for now
                decl = self.visit_bound_decls(section)
                task.setdefault('inputs', []).append(decl)
            # parameter_meta, and meta
            elif is_context(section, ('Parameter_meta_elementContext', 'Meta_elementContext')):
                pass
            else:
                raise RuntimeError(f'Unrecognized task element in visitTask(): {type(section)}')

    def visit_task_input(self, ctx):
        """
        Contains an array of 'any_decls', which can be unbound or bound declarations.
        Example:
            input {
              String in_str = "twenty"
              Int in_int
            }

        Returns a list of tuple=(name, type, expr)
        """
        return [self.visit_any_decls(decl) for decl in ctx.any_decls()]

    def visit_task_output(self, ctx):
        """
        Contains an array of 'bound_decls' (unbound_decls not allowed).
        Example:
            output {
              String out_str = read_string(stdout())
            }

        Returns a list of tuple=(name, type, expr)
        """
        return [self.visit_bound_decls(decl) for decl in ctx.bound_decls()]

    def visit_task_command(self, ctx):
        """
        Parses the command section of the WDL task.

        Contains a `string_part` plus any number of `expr_with_string`s.
        The following example command:
            'echo ${var1} ${var2} > output_file.txt'
        Has 3 parts:
                1. string_part: 'echo '
                2. expr_with_string, which has two parts:
                        - expr_part: 'var1'
                        - string_part: ' '
                1. expr_with_string, which has two parts:
                        - expr_part: 'var2'
                        - string_part: ' > output_file.txt'

        Returns a list=[] of strings representing the parts of the command.
            e.g. [string_part, expr_part, string_part, ...]
        """
        parts = []

        # add the first part
        str_part = self.visit_task_command_string_part(ctx.task_command_string_part())
        if str_part:
            parts.append(f"r'''{str_part}'''")

        # add the rest
        for group in ctx.task_command_expr_with_string():
            expr_part, str_part = self.visit_task_command_expr_with_string(group)
            parts.append(expr_part)
            if str_part:
                parts.append(f"r'''{str_part}'''")

        return parts

    def visit_task_command_string_part(self, ctx):
        """
        Returns a string representing the string_part.
        """
        # join here because a string that contains '$', '{', or '}' is split
        return ''.join(part.getText() for part in ctx.CommandStringPart())

    def visit_task_command_expr_with_string(self, ctx):
        """
        Returns a tuple=(expr_part, string_part).
        """
        return (self.visit_task_command_expr_part(ctx.task_command_expr_part()),
                self.visit_task_command_string_part(ctx.task_command_string_part()))

    def visit_task_command_expr_part(self, ctx):
        """
        Contains the expression inside ${expr}. Same function as `self.visit_string_expr_part()`.

        Returns the expression.
        """
        return self.visit_string_expr_part(ctx)

    def visit_task_runtime(self, ctx):
        """
        Contains an array of `task_runtime_kv`s.

        Returns a dict={key: value} where key can be 'docker', 'cpu', 'memory',
        'cores', or 'disks'.
        """
        return OrderedDict((kv.children[0].getText(),  # runtime key
                            self.visit_expr(kv.expr()))  # runtime value
                           for kv in ctx.task_runtime_kv())

    # Shared

    def visit_any_decls(self, ctx):
        """
        Contains a bound or unbound declaration.
        """
        if ctx.bound_decls():
            return self.visit_bound_decls(ctx.bound_decls())
        elif ctx.unbound_decls():
            return self.visit_unbound_decls(ctx.unbound_decls())
        else:
            raise RuntimeError(f'Unrecognized declaration: {type(ctx)}')

    def visit_unbound_decls(self, ctx):
        """
        Contains an unbound declaration. E.g.: `String in_str`.

        Returns a tuple=(name, type, expr), where `expr` is None.
        """
        name = ctx.Identifier().getText()
        type_ = self.visit_wdl_type(ctx.wdl_type())
        return name, type_, None

    def visit_bound_decls(self, ctx):
        """
        Contains a bound declaration. E.g.: `String in_str = "some string"`.

        Returns a tuple=(name, type, expr).
        """
        name = ctx.Identifier().getText()
        type_ = self.visit_wdl_type(ctx.wdl_type())
        expr = self.visit_expr(ctx.expr())

        return name, type_, expr

    def visit_wdl_type(self, ctx):
        """
        Returns a WDLType instance.
        """
        identifier = ctx.type_base().children[0]
        optional = ctx.OPTIONAL() is not None

        # primitives
        if is_context(identifier, 'TerminalNodeImpl'):

            # TODO: implement Object type
            return self.create_wdl_primitive_type(key=identifier.getText(), optional=optional)
        # compound types
        else:
            name = identifier.children[0].getText()  # the first child is the name of the type.
            type_ = identifier.wdl_type()
            if isinstance(type_, list):
                elements = [self.visit_wdl_type(element) for element in type_]
            else:
                elements = [self.visit_wdl_type(type_)]
            return self.create_wdl_compound_type(key=name, elements=elements, optional=optional)

    def visit_primitive_literal(self, ctx):
        """
        Returns the primitive literal as a string.
        """
        is_bool = ctx.BoolLiteral()
        if is_bool:
            val = is_bool.getText()
            if val not in ('true', 'false'):
                raise TypeError(f'Parsed boolean ({val}) must be expressed as "true" or "false".')
            return val.capitalize()
        elif is_context(ctx.children[0], 'StringContext'):
            return self.visit_string(ctx.children[0])
        elif is_context(ctx.children[0], ('TerminalNodeImpl',  # this also includes variables
                                          'NumberContext')):
            return ctx.children[0].getText()
        else:
            raise RuntimeError(f'Primitive literal has unknown child: {type(ctx.children[0])}.')

    def visit_number(self, ctx):
        """
        Contains an `IntLiteral` or a `FloatLiteral`.
        """
        return ctx.children[0].getText()

    def visit_string(self, ctx):
        """
        Contains a `string_part` followed by an array of `string_expr_with_string_part`s.
        """
        string = self.visit_string_part(ctx.string_part())

        for part in ctx.string_expr_with_string_part():
            string += f' + {self.visit_string_expr_with_string_part(part)}'

        return string

    def visit_string_expr_with_string_part(self, ctx):
        """
        Contains a `string_expr_part` and a `string_part`.
        """
        expr = self.visit_string_expr_part(ctx.string_expr_part())
        part = self.visit_string_part(ctx.string_part())

        if not part:
            return expr

        return f'{expr} + {part}'

    def visit_string_expr_part(self, ctx):
        """
        Contains an array of `expression_placeholder_option`s and an `expr`.
        """
        # See https://github.com/openwdl/wdl/blob/main/versions/1.0/parsers/antlr4/WdlV1Parser.g4#L56

        options = {}

        for opt in ctx.expression_placeholder_option():
            key, val = self.visit_expression_placeholder_option(opt)
            options[key] = val

        expr = self.visit_expr(ctx.expr())

        if len(options) == 0:
            return expr
        elif 'sep' in options:
            sep = options['sep']
            return f'{sep}.join(str(x) for x in {expr})'
        elif 'default' in options:
            default = options['default']
            return f'({expr} if {expr} else {default})'
        else:
            raise NotImplementedError(options)

    def visit_string_part(self, ctx):
        """
        Returns a string representing the string_part.
        """
        # join here because a string that contains '$', '{', or '}' is split
        part = ''.join(part.getText() for part in ctx.StringPart())

        if part:
            return f"'{part}'"
        return None

    def visit_expression_placeholder_option(self, ctx):
        """
        Expression placeholder options.

        Can match one of the following:
              BoolLiteral EQUAL (string | number)
              DEFAULT EQUAL (string | number)
              SEP EQUAL (string | number)

        See https://github.com/openwdl/wdl/blob/main/versions/1.0/SPEC.md#expression-placeholder-options

        e.g.: ${sep=", " array_value}
        e.g.: ${true="--yes" false="--no" boolean_value}
        e.g.: ${default="foo" optional_value}

        Returns a tuple=(key, value)
        """
        assert len(ctx.children) == 3

        param = ctx.children[0].getText()
        str_or_num = ctx.children[2]
        val = self.visit_string(str_or_num) \
            if is_context(str_or_num, 'StringContext') else self.visit_number(str_or_num)

        return param, val

    def visit_expr(self, ctx):
        """
        Expression root.
        """
        return self.visit_infix0(ctx.expr_infix())

    def visit_infix0(self, ctx):
        """
        Expression infix0 (LOR).
        """
        infix = ctx.expr_infix0()
        if is_context(infix, 'LorContext'):
            return self.visit_lor(infix)
        return self.visit_infix1(infix)

    def visit_lor(self, ctx):
        """
        Logical OR expression.
        """
        lhs = self.visit_infix0(ctx)
        rhs = self.visit_infix1(ctx)
        return f'{lhs} or {rhs}'

    def visit_infix1(self, ctx):
        """
        Expression infix1 (LAND).
        """
        infix = ctx.expr_infix1()
        if is_context(infix, 'LandContext'):
            return self.visit_land(infix)
        return self.visit_infix2(infix)

    def visit_land(self, ctx):
        """
        Logical AND expresion.
        """
        lhs = self.visit_infix1(ctx)
        rhs = self.visit_infix2(ctx)
        return f'{lhs} and {rhs}'

    def visit_infix2(self, ctx):
        """
        Expression infix2 (comparisons).
        """
        infix = ctx.expr_infix2()
        if is_context(infix, 'EqeqContext'):
            return self._visit_infix2(infix, '==')
        elif is_context(infix, 'NeqContext'):
            return self._visit_infix2(infix, '!=')
        elif is_context(infix, 'LteContext'):
            return self._visit_infix2(infix, '<=')
        elif is_context(infix, 'GteContext'):
            return self._visit_infix2(infix, '>=')
        elif is_context(infix, 'LtContext'):
            return self._visit_infix2(infix, '<')
        elif is_context(infix, 'GtContext'):
            return self._visit_infix2(infix, '>')
        # continue down our path
        return self.visit_infix3(infix)

    def _visit_infix2(self, ctx, operation: str):
        """
        :param operation: Operation as a string.
        """
        lhs = self.visit_infix2(ctx)
        rhs = self.visit_infix3(ctx)
        return f'{lhs} {operation} {rhs}'

    def visit_infix3(self, ctx):
        """
        Expression infix3 (add/subtract).
        """
        infix = ctx.expr_infix3()
        if is_context(infix, 'AddContext'):
            return self._visit_infix3(infix, '+')
        elif is_context(infix, 'SubContext'):
            return self._visit_infix3(infix, '-')
        # continue down our path
        return self.visit_infix4(infix)

    def _visit_infix3(self, ctx, operation: str):
        """
        :param operation: Operation as a string.
        """
        lhs = self.visit_infix3(ctx)
        rhs = self.visit_infix4(ctx)
        return f'{lhs} {operation} {rhs}'

    def visit_infix4(self, ctx):
        """
        Expression infix4 (multiply/divide/modulo).
        """
        infix = ctx.expr_infix4()
        if is_context(infix, 'MulContext'):
            return self._visit_infix4(infix, '*')
        elif is_context(infix, 'DivideContext'):
            return self._visit_infix4(infix, '/')
        elif is_context(infix, 'ModContext'):
            return self._visit_infix4(infix, '%')
        # continue down our path
        return self.visit_infix5(infix)

    def _visit_infix4(self, ctx, operation: str):
        """
        :param operation: Operation as a string.
        """
        lhs = self.visit_infix4(ctx)
        rhs = self.visit_infix5(ctx)
        return f'{lhs} {operation} {rhs}'

    def visit_infix5(self, ctx):
        """
        Expression infix5.
        """
        return self.visit_expr_core(ctx.expr_infix5().expr_core())

    def visit_expr_core(self, expr):
        """
        Expression core.
        """
        # TODO: implement map_literal, object_literal, and left_name

        if is_context(expr, 'ApplyContext'):
            return self.visit_apply(expr)
        elif is_context(expr, 'Array_literalContext'):
            return self.visit_array_literal(expr)
        elif is_context(expr, 'Pair_literalContext'):
            return self.visit_pair_literal(expr)
        elif is_context(expr, 'IfthenelseContext'):
            return self.visit_ifthenelse(expr)
        elif is_context(expr, 'Expression_groupContext'):
            return self.visit_expression_group(expr)
        elif is_context(expr, 'AtContext'):
            return self.visit_at(expr)
        elif is_context(expr, 'Get_nameContext'):
            return self.visit_get_name(expr)
        elif is_context(expr, 'NegateContext'):
            return self.visit_negate(expr)
        elif is_context(expr, 'UnarysignedContext'):
            return self.visit_unarysigned(expr)
        elif is_context(expr, 'PrimitivesContext'):
            return self.visit_primitives(expr)

        raise NotImplementedError(f"Expression context '{type(expr)}' is not supported.")

    # expr_core

    def visit_apply(self, ctx):
        """
        A function call.
        Pattern: Identifier LPAREN (expr (COMMA expr)*)? RPAREN
        """
        fn = ctx.Identifier().getText()
        params = ', '.join(self.visit_expr(expr) for expr in ctx.expr())

        if fn == 'stdout':
            return '_toil_wdl_internal__stdout_file'
        elif fn == 'stderr':
            return '_toil_wdl_internal__stderr_file'
        elif fn in ('range', 'zip'):
            # replace python built-in functions
            return f'wdl_{fn}'

        call = f'{fn}({params}'

        # append necessary params for i/o functions
        if fn == 'glob':
            return call + ', tempDir)'
        elif fn == 'size':
            return call + (params + ', ' if params else '') + 'fileStore=fileStore)'
        elif fn in ('write_lines', 'write_tsv', 'write_json', 'write_map'):
            return call + ', temp_dir=tempDir, file_store=fileStore)'
        else:
            return call + ')'

    def visit_array_literal(self, ctx):
        """
        Pattern: LBRACK (expr (COMMA expr)*)* RBRACK
        """
        return f"[{', '.join(self.visit_expr(expr) for expr in ctx.expr())}]"

    def visit_pair_literal(self, ctx):
        """
        Pattern: LPAREN expr COMMA expr RPAREN
        """
        return f"({self.visit_expr(ctx.expr(0))}, {self.visit_expr(ctx.expr(1))})"

    def visit_ifthenelse(self, ctx):
        """
        Ternary expression.
        Pattern: IF expr THEN expr ELSE expr
        """
        if_true = self.visit_expr(ctx.expr(0))
        condition = self.visit_expr(ctx.expr(1))
        if_false = self.visit_expr(ctx.expr(2))

        return f'({condition} if {if_true} else {if_false})'

    def visit_expression_group(self, ctx):
        """
        Pattern: LPAREN expr RPAREN
        """
        return f'({self.visit_expr(ctx.expr())})'

    def visit_at(self, ctx):
        """
        Array or map lookup.
        Pattern: expr_core LBRACK expr RBRACK
        """
        expr_core = self.visit_expr_core(ctx.expr_core())
        expr = self.visit_expr(ctx.expr())

        # parenthesis must be removed because 'i[0]' works, but '(i)[0]' does not
        if expr_core[0] == '(' and expr_core[-1] == ')':
            expr_core = expr_core[1:-1]

        return f'{expr_core}[{expr}]'

    def visit_get_name(self, ctx):
        """
        Member access.
        Pattern: expr_core DOT Identifier
        """
        expr_core = self.visit_expr_core(ctx.expr_core())
        identifier = ctx.Identifier().getText()

        if identifier in ('left', 'right'):
            # hack-y way to make sure pair.left and pair.right are parsed correctly.
            return f'({expr_core}.{identifier})'

        return f'({expr_core}_{identifier})'

    def visit_negate(self, ctx):
        """
        Pattern: NOT expr
        """
        return f'(not {self.visit_expr(ctx.expr())})'

    def visit_unarysigned(self, ctx):
        """
        Pattern: (PLUS | MINUS) expr
        """
        plus: bool = ctx.PLUS() is not None
        expr = self.visit_expr(ctx.expr())

        if plus:
            return f'(+{expr})'
        return f'(-{expr})'

    def visit_primitives(self, ctx):
        """
        Expression alias for primitive literal.
        """
        return self.visit_primitive_literal(ctx.primitive_literal())



import json
import logging
import os
import time
import urllib.request
from urllib.error import HTTPError
from typing import Dict, Optional, Iterator, cast

from botocore.client import BaseClient

from toil.lib.retry import retry

logger = logging.getLogger(__name__)


def get_flatcar_ami(ec2_client: BaseClient, architecture: str = 'amd64') -> str:
    """
    Retrieve the flatcar AMI image to use as the base for all Toil autoscaling instances.

    AMI must be available to the user on AWS (attempting to launch will return a 403 otherwise).

    Priority is:
      1. User specified AMI via TOIL_AWS_AMI
      2. Official AMI from stable.release.flatcar-linux.net
      3. Search the AWS Marketplace

    If all of these sources fail, we raise an error to complain.

    :param ec2_client: Boto3 EC2 Client
    :param architecture: The architecture type for the new AWS machine. Can be either amd64 or arm64
    """

    # Take a user override
    ami = os.environ.get('TOIL_AWS_AMI')
    try_number = 0
    if not ami:
        logger.debug('No AMI found in TOIL_AWS_AMI; checking stable Flatcar release feed')
        ami = feed_flatcar_ami_release(ec2_client=ec2_client, architecture=architecture, source='stable')
    if not ami:
        logger.warning('No available AMI found in Flatcar release feed; checking marketplace')
        ami = aws_marketplace_flatcar_ami_search(ec2_client=ec2_client, architecture=architecture)
    if not ami:
        logger.debug('No AMI found in marketplace; checking Toil Flatcar release feed')
        ami = feed_flatcar_ami_release(ec2_client=ec2_client, architecture=architecture, source='toil')
    if not ami:
        logger.debug('No AMI found in Toil project feed; checking beta Flatcar release feed')
        ami = feed_flatcar_ami_release(ec2_client=ec2_client, architecture=architecture, source='beta')
    if not ami:
        logger.debug('No AMI found in beta Flatcar release feed; checking archived Flatcar release feed')
        ami = feed_flatcar_ami_release(ec2_client=ec2_client, architecture=architecture, source='archive')
    if not ami:
        logger.critical('No available Flatcar AMI in any source!')
        raise RuntimeError(f'Unable to fetch the latest flatcar image. Upload '
                           f'https://stable.release.flatcar-linux.net/{architecture}-usr/current/flatcar_production_ami_image.bin.bz2 '
                           f'to AWS as am AMI and set TOIL_AWS_AMI in the environment to its AMI ID.')
    logger.info('Selected Flatcar AMI: %s', ami)
    return ami

@retry(errors=[HTTPError])
def _fetch_flatcar_feed(architecture: str = 'amd64', source: str = 'stable') -> bytes:
    """
    Get the binary data of the Flatcar release feed for the given architecture.
    
    :param source: can be set to a Flatcar release channel ('stable', 'beta',
           or 'alpha'), 'archive' to check the Internet Archive for a feed,
           and 'toil' to check if the Toil project has put up a feed.
    
    :raises HTTPError: if the feed cannot be fetched.
    """
    
    # We have a few places we know to get the feed from.
    JSON_FEED_URL = {
        'stable': f'https://stable.release.flatcar-linux.net/{architecture}-usr/current/flatcar_production_ami_all.json',
        'beta': f'https://beta.release.flatcar-linux.net/{architecture}-usr/current/flatcar_production_ami_all.json',
        'alpha': f'https://alpha.release.flatcar-linux.net/{architecture}-usr/current/flatcar_production_ami_all.json',
        'archive': f'https://web.archive.org/web/20220625112618if_/https://stable.release.flatcar-linux.net/{architecture}-usr/current/flatcar_production_ami_all.json',
        'toil': f'https://raw.githubusercontent.com/DataBiosphere/toil/master/contrib/flatcar/{architecture}-usr/current/flatcar_production_ami_all.json'
    }[source]
    return cast(bytes, urllib.request.urlopen(JSON_FEED_URL).read())

def flatcar_release_feed_amis(region: str, architecture: str = 'amd64', source: str = 'stable') -> Iterator[str]:
    """
    Yield AMI IDs for the given architecture from the Flatcar release feed.
    
    :param source: can be set to a Flatcar release channel ('stable', 'beta',
           or 'alpha'), 'archive' to check the Internet Archive for a feed,
           and 'toil' to check if the Toil project has put up a feed.
    
    Retries if the release feed cannot be fetched. If the release feed has a
    permanent error, yields nothing. If some entries in the release feed are
    unparseable, yields the others.
    """
    
    # If we get non-JSON content we want to retry.
    MAX_TRIES = 3
    
    try_number = 0
    while try_number < MAX_TRIES:
        try:
            feed = json.loads(_fetch_flatcar_feed(architecture, source))
            break
        except HTTPError:
            # Flatcar servers did not return the feed
            logger.exception(f'Could not retrieve {source} Flatcar release feed JSON')
            # Don't retry
            return
        except json.JSONDecodeError:
            # Feed is not JSON
            logger.exception(f'Could not decode {source} Flatcar release feed JSON')
            # Try again
            try_number += 1
            continue
    if try_number == MAX_TRIES:
        # We could not get the JSON
        logger.error(f'Could not get a readable {source} Flatcar release feed JSON')
        # Bail on this method
        return

    for ami_record in feed.get('amis', []):
        # Scan the list of regions
        if ami_record.get('name', None) == region:
            # When we find ours, return the AMI ID
            if 'hvm' in ami_record:
                yield ami_record['hvm']
            # And stop, there should be one per region.
            return
    # We didn't find our region
    logger.warning(f'Flatcar {source} release feed does not have an image for region {region}')
    
    

@retry()  # TODO: What errors do we get for timeout, JSON parse failure, etc?
def feed_flatcar_ami_release(ec2_client: BaseClient, architecture: str = 'amd64', source: str = 'stable') -> Optional[str]:
    """
    Check a Flatcar release feed for the latest flatcar AMI.
    
    Verify it's on AWS.

    :param ec2_client: Boto3 EC2 Client
    :param architecture: The architecture type for the new AWS machine. Can be either amd64 or arm64
    :param source: can be set to a Flatcar release channel ('stable', 'beta',
           or 'alpha'), 'archive' to check the Internet Archive for a feed,
           and 'toil' to check if the Toil project has put up a feed.
    """
    # Flatcar images only live for 9 months.
    # Rather than hardcode a list of AMIs by region that will die, we use
    # their JSON feed of the current ones.

    region = ec2_client._client_config.region_name  # type: ignore 
                
    for ami in flatcar_release_feed_amis(region, architecture, source):
        # verify it exists on AWS
        response = ec2_client.describe_images(Filters=[{'Name': 'image-id', 'Values': [ami]}])  # type: ignore
        if len(response['Images']) == 1 and response['Images'][0]['State'] == 'available':
            return ami 
        else:
            logger.warning(f'Flatcar release feed suggests image {ami} which does not exist on AWS in {region}')
    # We didn't find it
    logger.warning(f'Flatcar release feed does not have an image for region {region} that exists on AWS')
    return None


@retry()  # TODO: What errors do we get for timeout, JSON parse failure, etc?
def aws_marketplace_flatcar_ami_search(ec2_client: BaseClient, architecture: str = 'amd64') -> Optional[str]:
    """Query AWS for all AMI names matching 'Flatcar-stable-*' and return the most recent one."""

    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_images
    # Possible arch choices on AWS: 'i386'|'x86_64'|'arm64'|'x86_64_mac'
    architecture_mapping = {'amd64': 'x86_64',
                            'arm64': 'arm64'}
    response: dict = ec2_client.describe_images(Owners=['aws-marketplace'],  # type: ignore
                                                Filters=[{'Name': 'name', 'Values': ['Flatcar-stable-*']}])
    latest: Dict[str, str] = {'CreationDate': '0lder than atoms.'}
    for image in response['Images']:
        if image['Architecture'] == architecture_mapping[architecture] and image['State'] == 'available':
            if image['CreationDate'] > latest['CreationDate']:
                latest = image
    return latest.get('ImageId', None)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import time
from abc import ABCMeta, abstractmethod
from datetime import datetime
from queue import Empty, Queue
from threading import Lock, Thread
from typing import Any, Dict, List, Optional, Tuple, Union

from toil.batchSystems.abstractBatchSystem import (BatchJobExitReason,
                                                   UpdatedBatchJobInfo)
from toil.batchSystems.cleanup_support import BatchSystemCleanupSupport
from toil.bus import ExternalBatchIdMessage
from toil.job import JobDescription, AcceleratorRequirement
from toil.lib.misc import CalledProcessErrorStderr

logger = logging.getLogger(__name__)


# Internally we throw around these flat tuples of random important things about a job.
# Assigned ID
# Required cores
# Required memory
# Command to run
# Unit name of the job
# Environment dict for the job
# Accelerator requirements for the job
JobTuple = Tuple[int, float, int, str, str, Dict[str, str], List[AcceleratorRequirement]]

class AbstractGridEngineBatchSystem(BatchSystemCleanupSupport):
    """
    A partial implementation of BatchSystemSupport for batch systems run on a
    standard HPC cluster. By default auto-deployment is not implemented.
    """

    class Worker(Thread, metaclass=ABCMeta):

        def __init__(self, newJobsQueue: Queue, updatedJobsQueue: Queue, killQueue: Queue, killedJobsQueue: Queue, boss: 'AbstractGridEngineBatchSystem') -> None:
            """
            Abstract worker interface class. All instances are created with five
            initial arguments (below). Note the Queue instances passed are empty.

            :param newJobsQueue: a Queue of new (unsubmitted) jobs
            :param updatedJobsQueue: a Queue of jobs that have been updated
            :param killQueue: a Queue of active jobs that need to be killed
            :param killedJobsQueue: Queue of killed jobs for this worker
            :param boss: the AbstractGridEngineBatchSystem instance that
                         controls this AbstractGridEngineWorker

            """
            Thread.__init__(self)
            self.boss = boss
            self.boss.config.statePollingWait = \
                self.boss.config.statePollingWait or self.boss.getWaitDuration()
            self.newJobsQueue = newJobsQueue
            self.updatedJobsQueue = updatedJobsQueue
            self.killQueue = killQueue
            self.killedJobsQueue = killedJobsQueue
            self.waitingJobs: List[JobTuple] = list()
            self.runningJobs = set()
            self.runningJobsLock = Lock()
            self.batchJobIDs: Dict[int, str] = dict()
            self._checkOnJobsCache = None
            self._checkOnJobsTimestamp = None

        def getBatchSystemID(self, jobID: int) -> str:
            """
            Get batch system-specific job ID

            Note: for the moment this is the only consistent way to cleanly get
            the batch system job ID

            :param jobID: Toil BatchSystem numerical job ID
            """
            if jobID not in self.batchJobIDs:
                raise RuntimeError("Unknown jobID, could not be converted")

            (job, task) = self.batchJobIDs[jobID]
            if task is None:
                return str(job)
            else:
                return str(job) + "." + str(task)

        def forgetJob(self, jobID: int) -> None:
            """
            Remove jobID passed

            :param jobID: toil job ID
            """
            with self.runningJobsLock:
                self.runningJobs.remove(jobID)
            del self.batchJobIDs[jobID]

        def createJobs(self, newJob: JobTuple) -> bool:
            """
            Create a new job with the given attributes.

            Implementation-specific; called by AbstractGridEngineWorker.run()
            """
            activity = False
            # Load new job id if present:
            if newJob is not None:
                self.waitingJobs.append(newJob)
            # Launch jobs as necessary:
            while len(self.waitingJobs) > 0 and \
                    len(self.runningJobs) < int(self.boss.config.max_jobs):
                activity = True
                jobID, cpu, memory, command, jobName, environment, gpus = self.waitingJobs.pop(0)

                # prepare job submission command
                subLine = self.prepareSubmission(cpu, memory, jobID, command, jobName, environment, gpus)
                logger.debug("Running %r", subLine)
                batchJobID = self.boss.with_retries(self.submitJob, subLine)
                if self.boss._outbox is not None:
                    #JobID corresponds to the toil version of the jobID, dif from jobstore idea of the id, batchjobid is what we get from slurm
                    self.boss._outbox.publish(ExternalBatchIdMessage(jobID, batchJobID, self.boss.__class__.__name__))

                logger.debug("Submitted job %s", str(batchJobID))

                # Store dict for mapping Toil job ID to batch job ID
                # TODO: Note that this currently stores a tuple of (batch system
                # ID, Task), but the second value is None by default and doesn't
                # seem to be used
                self.batchJobIDs[jobID] = (batchJobID, None)

                # Add to queue of running jobs
                with self.runningJobsLock:
                    self.runningJobs.add(jobID)

            return activity

        def killJobs(self):
            """
            Kill any running jobs within worker
            """
            killList = list()
            while True:
                try:
                    jobId = self.killQueue.get(block=False)
                except Empty:
                    break
                else:
                    killList.append(jobId)

            if not killList:
                return False

            # Do the dirty job
            for jobID in list(killList):
                if jobID in self.runningJobs:
                    logger.debug('Killing job: %s', jobID)

                    # this call should be implementation-specific, all other
                    # code is redundant w/ other implementations
                    self.killJob(jobID)
                else:
                    if jobID in self.waitingJobs:
                        self.waitingJobs.remove(jobID)
                    self.killedJobsQueue.put(jobID)
                    killList.remove(jobID)

            # Wait to confirm the kill
            while killList:
                for jobID in list(killList):
                    batchJobID = self.getBatchSystemID(jobID)
                    if self.boss.with_retries(self.getJobExitCode, batchJobID) is not None:
                        logger.debug('Adding jobID %s to killedJobsQueue', jobID)
                        self.killedJobsQueue.put(jobID)
                        killList.remove(jobID)
                        self.forgetJob(jobID)
                if len(killList) > 0:
                    logger.warning("Some jobs weren't killed, trying again in %is.", self.boss.sleepSeconds())

            return True

        def checkOnJobs(self):
            """Check and update status of all running jobs.

            Respects statePollingWait and will return cached results if not within
            time period to talk with the scheduler.
            """

            if self._checkOnJobsTimestamp:
                time_since_last_check = (datetime.now() - self._checkOnJobsTimestamp).total_seconds()
                if time_since_last_check < self.boss.config.statePollingWait:
                    return self._checkOnJobsCache

            activity = False
            running_job_list = list(self.runningJobs)
            batch_job_id_list = [self.getBatchSystemID(j) for j in running_job_list]
            if batch_job_id_list:
                try:
                    # Get the statuses as a batch
                    statuses = self.boss.with_retries(
                        self.coalesce_job_exit_codes, batch_job_id_list
                    )
                except NotImplementedError:
                    # We have to get the statuses individually
                    for running_job_id, batch_job_id in zip(running_job_list, batch_job_id_list):
                        status = self.boss.with_retries(self.getJobExitCode, batch_job_id)
                        activity = self._handle_job_status(
                            running_job_id, status, activity
                        )
                else:
                    # We got the statuses as a batch
                    for running_job_id, status in zip(running_job_list, statuses):
                        activity = self._handle_job_status(
                            running_job_id, status, activity
                        )

            self._checkOnJobsCache = activity
            self._checkOnJobsTimestamp = datetime.now()
            return activity

        def _handle_job_status(
            self, job_id: int, status: Union[int, None], activity: bool
        ) -> bool:
            """
            Helper method for checkOnJobs to handle job statuses
            """
            if status is not None:
                self.updatedJobsQueue.put(
                    UpdatedBatchJobInfo(
                        jobID=job_id, exitStatus=status, exitReason=None, wallTime=None
                    )
                )
                self.forgetJob(job_id)
                return True
            if status is not None and isinstance(status, BatchJobExitReason):
                self.updatedJobsQueue.put(
                    UpdatedBatchJobInfo(
                        jobID=job_id, exitStatus=1, exitReason=status, wallTime=None
                    )
                )
                self.forgetJob(job_id)
                return True
            return activity

        def _runStep(self):
            """return True if more jobs, False is all done"""
            activity = False
            newJob = None
            if not self.newJobsQueue.empty():
                activity = True
                newJob = self.newJobsQueue.get()
                if newJob is None:
                    logger.debug('Received queue sentinel.')
                    return False
            if self.killJobs():
                activity = True
            if self.createJobs(newJob):
                activity = True
            if self.checkOnJobs():
                activity = True
            if not activity:
                logger.debug('No activity, sleeping for %is', self.boss.sleepSeconds())
            return True

        def run(self):
            """
            Run any new jobs
            """
            try:
                while self._runStep():
                    pass
            except Exception as ex:
                logger.error("GridEngine like batch system failure", exc_info=ex)
                raise

        def coalesce_job_exit_codes(self, batch_job_id_list: list) -> list:
            """
            Returns exit codes for a list of jobs.

            Called by AbstractGridEngineWorker.checkOnJobs().

            This is an optional part of the interface. It should raise
            NotImplementedError if not actually implemented for a particular
            scheduler.

            :param string batch_job_id_list: List of batch system job ID
            """
            raise NotImplementedError()

        @abstractmethod
        def prepareSubmission(self,
                              cpu: int,
                              memory: int,
                              jobID: int,
                              command: str,
                              jobName: str,
                              job_environment: Optional[Dict[str, str]] = None,
                              gpus: Optional[int] = None) -> List[str]:
            """
            Preparation in putting together a command-line string
            for submitting to batch system (via submitJob().)

            :param: int cpu
            :param: int memory
            :param: int jobID: Toil job ID
            :param: string subLine: the command line string to be called
            :param: string jobName: the name of the Toil job, to provide metadata to batch systems if desired
            :param: dict job_environment: the environment variables to be set on the worker

            :rtype: List[str]
            """
            raise NotImplementedError()

        @abstractmethod
        def submitJob(self, subLine):
            """
            Wrapper routine for submitting the actual command-line call, then
            processing the output to get the batch system job ID

            :param: string subLine: the literal command line string to be called

            :rtype: string: batch system job ID, which will be stored internally
            """
            raise NotImplementedError()

        @abstractmethod
        def getRunningJobIDs(self):
            """
            Get a list of running job IDs. Implementation-specific; called by boss
            AbstractGridEngineBatchSystem implementation via
            AbstractGridEngineBatchSystem.getRunningBatchJobIDs()

            :rtype: list
            """
            raise NotImplementedError()

        @abstractmethod
        def killJob(self, jobID):
            """
            Kill specific job with the Toil job ID. Implementation-specific; called
            by AbstractGridEngineWorker.killJobs()

            :param string jobID: Toil job ID
            """
            raise NotImplementedError()

        @abstractmethod
        def getJobExitCode(self, batchJobID):
            """
            Returns job exit code or an instance of abstractBatchSystem.BatchJobExitReason.
            if something else happened other than the job exiting.
            Implementation-specific; called by AbstractGridEngineWorker.checkOnJobs()

            :param string batchjobID: batch system job ID

            :rtype: int|toil.batchSystems.abstractBatchSystem.BatchJobExitReason: exit code int
                    or BatchJobExitReason if something else happened other than job exiting.
            """
            raise NotImplementedError()

    def __init__(self, config, maxCores, maxMemory, maxDisk):
        super().__init__(
            config, maxCores, maxMemory, maxDisk)
        self.config = config

        self.currentJobs = set()

        self.newJobsQueue = Queue()
        self.updatedJobsQueue = Queue()
        self.killQueue = Queue()
        self.killedJobsQueue = Queue()
        # get the associated worker class here
        self.worker = self.Worker(self.newJobsQueue, self.updatedJobsQueue,
                                  self.killQueue, self.killedJobsQueue, self)
        self.worker.start()
        self._getRunningBatchJobIDsTimestamp = None
        self._getRunningBatchJobIDsCache = {}

    @classmethod
    def supportsWorkerCleanup(cls):
        return False

    @classmethod
    def supportsAutoDeployment(cls):
        return False

    def issueBatchJob(self, jobDesc, job_environment: Optional[Dict[str, str]] = None):
        # Avoid submitting internal jobs to the batch queue, handle locally
        localID = self.handleLocalJob(jobDesc)
        if localID is not None:
            return localID
        else:
            self.check_resource_request(jobDesc)
            jobID = self.getNextJobID()
            self.currentJobs.add(jobID)
            gpus = 0
            if isinstance(jobDesc.accelerators, list):
                for accelerator in jobDesc.accelerators:
                    if accelerator['kind'] == 'gpu':
                        gpus = accelerator['count']
            else:
                gpus = jobDesc.accelerators

            self.newJobsQueue.put((jobID, jobDesc.cores, jobDesc.memory, jobDesc.command, jobDesc.get_job_kind(),
                                   job_environment, gpus))
            logger.debug("Issued the job command: %s with job id: %s and job name %s", jobDesc.command, str(jobID),
                         jobDesc.get_job_kind())
        return jobID

    def killBatchJobs(self, jobIDs):
        """
        Kills the given jobs, represented as Job ids, then checks they are dead by checking
        they are not in the list of issued jobs.
        """
        self.killLocalJobs(jobIDs)
        jobIDs = set(jobIDs)
        logger.debug('Jobs to be killed: %r', jobIDs)
        for jobID in jobIDs:
            self.killQueue.put(jobID)
        while jobIDs:
            killedJobId = self.killedJobsQueue.get()
            if killedJobId is None:
                break
            jobIDs.remove(killedJobId)
            if killedJobId in self._getRunningBatchJobIDsCache:
                # Running batch id cache can sometimes contain a job we kill, so to ensure cache doesn't contain the job, we delete it here
                del self._getRunningBatchJobIDsCache[killedJobId]
            if killedJobId in self.currentJobs:
                self.currentJobs.remove(killedJobId)
            if jobIDs:
                logger.debug('Some kills (%s) still pending, sleeping %is', len(jobIDs),
                             self.sleepSeconds())

    def getIssuedBatchJobIDs(self):
        """
        Gets the list of issued jobs
        """
        return list(self.getIssuedLocalJobIDs()) + list(self.currentJobs)

    def getRunningBatchJobIDs(self):
        """
        Retrieve running job IDs from local and batch scheduler.

        Respects statePollingWait and will return cached results if not within
        time period to talk with the scheduler.
        """
        if (self._getRunningBatchJobIDsTimestamp and (
                datetime.now() -
                self._getRunningBatchJobIDsTimestamp).total_seconds() <
                self.config.statePollingWait):
            batchIds = self._getRunningBatchJobIDsCache
        else:
            batchIds = self.with_retries(self.worker.getRunningJobIDs)
            self._getRunningBatchJobIDsCache = batchIds
            self._getRunningBatchJobIDsTimestamp = datetime.now()
        batchIds.update(self.getRunningLocalJobIDs())
        return batchIds

    def getUpdatedBatchJob(self, maxWait):
        local_tuple = self.getUpdatedLocalJob(0)
        if local_tuple:
            return local_tuple
        else:
            try:
                item = self.updatedJobsQueue.get(timeout=maxWait)
            except Empty:
                return None
            logger.debug('UpdatedJobsQueue Item: %s', item)
            self.currentJobs.remove(item.jobID)
            return item

    def shutdown(self) -> None:
        """
        Signals worker to shutdown (via sentinel) then cleanly joins the thread
        """
        self.shutdownLocal()
        newJobsQueue = self.newJobsQueue
        self.newJobsQueue = None

        newJobsQueue.put(None)
        self.worker.join()

    def setEnv(self, name, value=None):
        if value and ',' in value:
            raise ValueError(type(self).__name__ + " does not support commata in environment variable values")
        return super().setEnv(name, value)

    @classmethod
    def getWaitDuration(self):
        return 1

    def sleepSeconds(self, sleeptime=1):
        """ Helper function to drop on all state-querying functions to avoid over-querying.
        """
        time.sleep(sleeptime)
        return sleeptime

    def with_retries(self, operation, *args, **kwargs):
        """
        Call operation with args and kwargs. If one of the calls to an SGE
        command fails, sleep and try again for a set number of times.
        """
        maxTries = 3
        tries = 0
        while True:
            tries += 1
            try:
                return operation(*args, **kwargs)
            except CalledProcessErrorStderr as err:
                if tries < maxTries:
                    logger.error("Will retry errored operation %s, code %d: %s",
                                 operation.__name__, err.returncode, err.stderr)
                    time.sleep(self.config.statePollingWait)
                else:
                    logger.error("Failed operation %s, code %d: %s",
                                 operation.__name__, err.returncode, err.stderr)
                    raise err



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import gzip
import json
import logging
import os
import time
from argparse import ArgumentParser, Namespace
from logging.handlers import RotatingFileHandler
from threading import Event, Thread
from typing import IO, TYPE_CHECKING, Any, Callable, List, Optional, Union

from toil.lib.expando import Expando
from toil.lib.resources import get_total_cpu_time

if TYPE_CHECKING:
    from toil.common import Config
    from toil.jobStores.abstractJobStore import AbstractJobStore

logger = logging.getLogger(__name__)
root_logger = logging.getLogger()
toil_logger = logging.getLogger('toil')

DEFAULT_LOGLEVEL = logging.INFO
__loggingFiles = []


class StatsAndLogging:
    """A thread to aggregate statistics and logging."""
    def __init__(self, jobStore: 'AbstractJobStore', config: 'Config') -> None:
        self._stop = Event()
        self._worker = Thread(target=self.statsAndLoggingAggregator,
                              args=(jobStore, self._stop, config),
                              daemon=True)

    def start(self) -> None:
        """Start the stats and logging thread."""
        self._worker.start()

    @classmethod
    def formatLogStream(cls, stream: Union[IO[str], IO[bytes]], job_name: Optional[str] = None) -> str:
        """
        Given a stream of text or bytes, and the job name, job itself, or some
        other optional stringifyable identity info for the job, return a big
        text string with the formatted job log, suitable for printing for the
        user.

        We don't want to prefix every line of the job's log with our own
        logging info, or we get prefixes wider than any reasonable terminal
        and longer than the messages.

        :param stream: The stream of text or bytes to print for the user.
        """
        lines = [f'Log from job "{job_name}" follows:', '=========>']

        for line in stream:
            if isinstance(line, bytes):
                line = line.decode('utf-8', errors='replace')
            lines.append('\t' + line.rstrip('\n'))

        lines.append('<=========')

        return '\n'.join(lines)


    @classmethod
    def logWithFormatting(cls, jobStoreID: str, jobLogs: Union[IO[str], IO[bytes]], method: Callable[[str], None] = logger.debug,
                            message: Optional[str] = None) -> None:
        if message is not None:
            method(message)

        # Format and log the logs, identifying the job with its job store ID.
        method(cls.formatLogStream(jobLogs, jobStoreID))

    @classmethod
    def writeLogFiles(cls, jobNames: List[str], jobLogList: List[str], config: 'Config', failed: bool = False) -> None:
        def createName(logPath: str, jobName: str, logExtension: str, failed: bool = False) -> str:
            logName = jobName.replace('-', '--')
            logName = logName.replace('/', '-')
            logName = logName.replace(' ', '_')
            logName = logName.replace("'", '')
            logName = logName.replace('"', '')
            # Add a "failed_" prefix to logs from failed jobs.
            logName = ('failed_' if failed else '') + logName
            counter = 0
            while True:
                suffix = str(counter).zfill(3) + logExtension
                fullName = os.path.join(logPath, logName + suffix)
                #  The maximum file name size in the default HFS+ file system is 255 UTF-16 encoding units, so basically 255 characters
                if len(fullName) >= 255:
                    return fullName[:(255-len(suffix))] + suffix
                if not os.path.exists(fullName):
                    return fullName
                counter += 1

        mainFileName = jobNames[0]
        extension = '.log'
        writeFn: Callable[..., Any]
        if config.writeLogs:
            path = config.writeLogs
            writeFn = open
        elif config.writeLogsGzip:
            path = config.writeLogsGzip
            writeFn = gzip.open
            extension += '.gz'
        else:
            # we don't have anywhere to write the logs, return now
            return

        fullName = createName(path, mainFileName, extension, failed)
        with writeFn(fullName, 'wb') as f:
            for l in jobLogList:
                if isinstance(l, bytes):
                    l = l.decode('utf-8')
                if not l.endswith('\n'):
                    l += '\n'
                f.write(l.encode('utf-8'))
        for alternateName in jobNames[1:]:
            # There are chained jobs in this output - indicate this with a symlink
            # of the job's name to this file
            name = createName(path, alternateName, extension, failed)
            if not os.path.exists(name):
                os.symlink(os.path.relpath(fullName, path), name)

    @classmethod
    def statsAndLoggingAggregator(cls, jobStore: 'AbstractJobStore', stop: Event, config: 'Config') -> None:
        """
        The following function is used for collating stats/reporting log messages from the workers.
        Works inside of a thread, collates as long as the stop flag is not True.
        """
        #  Overall timing
        startTime = time.time()
        startClock = get_total_cpu_time()

        def callback(fileHandle: Union[IO[bytes], IO[str]]) -> None:
            statsStr = fileHandle.read()
            if not isinstance(statsStr, str):
                statsStr = statsStr.decode()
            stats = json.loads(statsStr, object_hook=Expando)
            try:
                logs = stats.workers.logsToMaster
            except AttributeError:
                # To be expected if there were no calls to logToMaster()
                pass
            else:
                for message in logs:
                    logger.log(int(message.level),
                               'Got message from job at time %s: %s',
                               time.strftime('%m-%d-%Y %H:%M:%S'), message.text)
            try:
                logs = stats.logs
            except AttributeError:
                pass
            else:
                # we may have multiple jobs per worker
                jobNames = logs.names
                messages = logs.messages
                cls.logWithFormatting(jobNames[0], messages,
                                      message='Received Toil worker log. Disable debug level logging to hide this output')
                cls.writeLogFiles(jobNames, messages, config=config)

        while True:
            # This is a indirect way of getting a message to the thread to exit
            if stop.is_set():
                jobStore.read_logs(callback)
                break
            if jobStore.read_logs(callback) == 0:
                time.sleep(0.5)  # Avoid cycling too fast

        # Finish the stats file
        text = json.dumps(dict(total_time=str(time.time() - startTime),
                               total_clock=str(get_total_cpu_time() - startClock)), ensure_ascii=True)
        jobStore.write_logs(text)

    def check(self) -> None:
        """
        Check on the stats and logging aggregator.
        :raise RuntimeError: If the underlying thread has quit.
        """
        if not self._worker.is_alive():
            raise RuntimeError("Stats and logging thread has quit")

    def shutdown(self) -> None:
        """Finish up the stats/logging aggregation thread."""
        logger.debug('Waiting for stats and logging collator thread to finish ...')
        startTime = time.time()
        self._stop.set()
        self._worker.join()
        logger.debug('... finished collating stats and logs. Took %s seconds', time.time() - startTime)
        # in addition to cleaning on exceptions, onError should clean if there are any failed jobs


def set_log_level(level: str, set_logger: Optional[logging.Logger] = None) -> None:
    """Sets the root logger level to a given string level (like "INFO")."""
    level = "CRITICAL" if level.upper() == "OFF" else level.upper()
    set_logger = set_logger if set_logger else root_logger
    set_logger.setLevel(level)

    # Suppress any random loggers introduced by libraries we use.
    # Especially boto/boto3.  They print too much.  -__-
    suppress_exotic_logging(__name__)


def add_logging_options(parser: ArgumentParser) -> None:
    """Add logging options to set the global log level."""
    group = parser.add_argument_group("Logging Options")
    default_loglevel = logging.getLevelName(DEFAULT_LOGLEVEL)

    levels = ['Critical', 'Error', 'Warning', 'Debug', 'Info']
    for level in levels:
        group.add_argument(f"--log{level}", dest="logLevel", default=default_loglevel, action="store_const",
                           const=level, help=f"Turn on loglevel {level}.  Default: {default_loglevel}.")

    levels += [l.lower() for l in levels] + [l.upper() for l in levels]
    group.add_argument("--logOff", dest="logLevel", default=default_loglevel,
                       action="store_const", const="CRITICAL", help="Same as --logCRITICAL.")
    group.add_argument("--logLevel", dest="logLevel", default=default_loglevel, choices=levels,
                       help=f"Set the log level. Default: {default_loglevel}.  Options: {levels}.")
    group.add_argument("--logFile", dest="logFile", help="File to log in.")
    group.add_argument("--rotatingLogging", dest="logRotating", action="store_true", default=False,
                       help="Turn on rotating logging, which prevents log files from getting too big.")


def configure_root_logger() -> None:
    """
    Set up the root logger with handlers and formatting.

    Should be called before any entry point tries to log anything,
    to ensure consistent formatting.
    """
    logging.basicConfig(format='[%(asctime)s] [%(threadName)-10s] [%(levelname).1s] [%(name)s] %(message)s',
                        datefmt='%Y-%m-%dT%H:%M:%S%z')
    root_logger.setLevel(DEFAULT_LOGLEVEL)


def log_to_file(log_file: Optional[str], log_rotation: bool) -> None:
    if log_file and log_file not in __loggingFiles:
        logger.debug(f"Logging to file '{log_file}'.")
        __loggingFiles.append(log_file)
        handler: Union[RotatingFileHandler, logging.FileHandler]
        if log_rotation:
            handler = RotatingFileHandler(log_file, maxBytes=1000000, backupCount=1)
        else:
            handler = logging.FileHandler(log_file)
        root_logger.addHandler(handler)


def set_logging_from_options(options: Union["Config", Namespace]) -> None:
    configure_root_logger()
    options.logLevel = options.logLevel or logging.getLevelName(root_logger.getEffectiveLevel())
    set_log_level(options.logLevel)
    logger.debug(f"Root logger is at level '{logging.getLevelName(root_logger.getEffectiveLevel())}', "
                 f"'toil' logger at level '{logging.getLevelName(toil_logger.getEffectiveLevel())}'.")

    # start logging to log file if specified
    log_to_file(options.logFile, options.logRotating)


def suppress_exotic_logging(local_logger: str) -> None:
    """
    Attempts to suppress the loggers of all non-Toil packages by setting them to CRITICAL.

    For example: 'requests_oauthlib', 'google', 'boto', 'websocket', 'oauthlib', etc.

    This will only suppress loggers that have already been instantiated and can be seen in the environment,
    except for the list declared in "always_suppress".

    This is important because some packages, particularly boto3, are not always instantiated yet in the
    environment when this is run, and so we create the logger and set the level preemptively.
    """
    never_suppress = ['toil', '__init__', '__main__', 'toil-rt', 'cwltool']
    always_suppress = ['boto3', 'boto', 'botocore']  # ensure we suppress even before instantiated

    top_level_loggers: List[str] = []

    # Due to https://stackoverflow.com/questions/61683713
    for pkg_logger in list(logging.Logger.manager.loggerDict.keys()) + always_suppress:
        if pkg_logger != local_logger:
            # many sub-loggers may exist, like "boto.a", "boto.b", "boto.c"; we only want the top_level: "boto"
            top_level_logger = pkg_logger.split('.')[0] if '.' in pkg_logger else pkg_logger

            if top_level_logger not in top_level_loggers + never_suppress:
                top_level_loggers.append(top_level_logger)
                logging.getLogger(top_level_logger).setLevel(logging.CRITICAL)
    logger.debug(f'Suppressing the following loggers: {set(top_level_loggers)}')



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import fcntl
import json
import logging
import multiprocessing
import os
import signal
import subprocess
import sys
import tempfile
import zipfile
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urldefrag

from celery.exceptions import SoftTimeLimitExceeded  # type: ignore

import toil.server.wes.amazon_wes_utils as amazon_wes_utils
from toil.common import Toil
from toil.jobStores.utils import generate_locator
from toil.server.celery_app import celery
from toil.server.utils import (WorkflowStateMachine,
                               connect_to_workflow_state_store,
                               download_file_from_internet,
                               download_file_from_s3,
                               get_file_class,
                               get_iso_time,
                               link_file)

logger = logging.getLogger(__name__)

# How many seconds should we give a Toil workflow to gracefully shut down
# before we kill it?
# Ought to be long enough to let it clean up its job store, but shorter than
# our patience for CANCELING WES workflows to time out to CANCELED.
WAIT_FOR_DEATH_TIMEOUT = 20

class ToilWorkflowRunner:
    """
    A class to represent a workflow runner to run the requested workflow.

    Responsible for parsing the user request into a shell command, executing
    that command, and collecting the outputs of the resulting workflow run.
    """

    def __init__(self, base_scratch_dir: str, state_store_url: str, workflow_id: str, request: Dict[str, Any], engine_options: List[str]):
        """
        Make a new ToilWorkflowRunner to actually run a workflow leader based
        on a WES request.

        :param base_scratch_dir: Base work directory. Workflow scratch directory
               will be in here under the workflow ID.
        :param state_store_url: URL to the state store through which we will
               communicate about workflow state with the WES server.
        :param workflow_id: ID of the workflow run.
        :param request: WES request information.
        :param engine_options: Extra options to pass to Toil.
        """

        # Find the scratch directory inside the base scratch directory
        self.scratch_dir = os.path.join(base_scratch_dir, workflow_id)

        # Connect to the workflow state store
        self.store = connect_to_workflow_state_store(state_store_url, workflow_id)
        # And use a state machine over that to look at workflow state
        self.state_machine = WorkflowStateMachine(self.store)

        self.request = request
        self.engine_options = engine_options

        self.wf_type: str = request["workflow_type"].lower().strip()
        self.version: str = request["workflow_type_version"]

        self.exec_dir = os.path.join(self.scratch_dir, "execution")
        self.out_dir = os.path.join(self.scratch_dir, "outputs")

        # Compose the right kind of job store to use it the user doesn't specify one.
        default_type = os.getenv('TOIL_WES_JOB_STORE_TYPE', 'file')
        self.default_job_store = generate_locator(default_type, local_suggestion=os.path.join(self.scratch_dir, "toil_job_store"))

        self.job_store = self.default_job_store

    def write_scratch_file(self, filename: str, contents: str) -> None:
        """
        Write a file to the scratch directory.
        """
        with open(os.path.join(self.scratch_dir, filename), "w") as f:
            f.write(contents)

    def get_state(self) -> str:
        return self.state_machine.get_current_state()

    def write_workflow(self, src_url: str) -> str:
        """
        Fetch the workflow file from its source and write it to a destination
        file.
        """
        logger.info(f"Processing workflow_url: '{src_url}'...")
        dest = os.path.join(self.exec_dir, os.path.basename(src_url))

        if src_url.startswith("file://"):
            logger.info("Linking workflow from filesystem.")
            link_file(src=src_url[7:], dest=dest)
        elif src_url.startswith(("http://", "https://")):
            logger.info("Downloading workflow_url from the Internet.")
            download_file_from_internet(src=src_url, dest=dest, content_type="text/")
        elif src_url.startswith("s3://"):
            logger.info("Downloading workflow_url from Amazon S3.")
            download_file_from_s3(src=src_url, dest=dest)
        else:
            logger.info("Using workflow from relative URL.")
            dest = os.path.join(self.exec_dir, src_url)

        # Make sure that the destination file actually exists
        if not os.path.isfile(urldefrag(dest)[0]):
            raise RuntimeError(f"Cannot resolve workflow file from: '{src_url}'.")

        return dest

    def sort_options(self, workflow_engine_parameters: Optional[Dict[str, Optional[str]]] = None) -> List[str]:
        """
        Sort the command line arguments in the order that can be recognized by
        the workflow execution engine.

        :param workflow_engine_parameters: User-specified parameters for this
        particular workflow. Keys are command-line options, and values are
        option arguments, or None for options that are flags.
        """
        options = []

        # First, we pass the default engine parameters
        options.extend(self.engine_options)

        if workflow_engine_parameters:
            # Then, we pass the user options specific for this workflow run.
            # This should override the default
            for key, value in workflow_engine_parameters.items():
                if value is None:  # flags
                    options.append(key)
                else:
                    options.append(f"{key}={value}")

        # We want to clean always by default, unless a particular job store or
        # a clean option was passed.
        clean = None

        # Parse options and drop options we may need to override.
        for option in options:
            if option.startswith("--jobStore="):
                self.job_store = option[11:]
                options.remove(option)
            if option.startswith(("--outdir=", "-o=")):
                # We need to generate this one ourselves.
                options.remove(option)
            if option.startswith("--clean="):
                clean = option[8:]

        cloud = False
        job_store_type, _ = Toil.parseLocator(self.job_store)
        if job_store_type in ("aws", "google", "azure"):
            cloud = True

        if self.job_store == self.default_job_store and clean is None:
            # User didn't specify a clean option, and we're on a default,
            # randomly generated job store, so we should clean it up even if we
            # crash.
            options.append("--clean=always")

        if self.wf_type in ("cwl", "wdl"):
            if not cloud:
                options.append("--outdir=" + self.out_dir)
            options.append("--jobStore=" + self.job_store)
        else:
            # TODO: find a way to communicate the out_dir to the Toil workflow.

            # append the positional jobStore argument at the end for Toil workflows
            options.append(self.job_store)

        return options

    def initialize_run(self) -> List[str]:
        """
        Write workflow and input files and construct a list of shell commands
        to be executed. Return that list of shell commands that should be
        executed in order to complete this workflow run.
        """
        # Obtain CWL-style workflow parameters from the request. Default to an
        # empty dict if not found, because we want to tolerate omitting this.
        workflow_params = self.request.get("workflow_params", {})

        # And any workflow engine parameters the user specified.
        workflow_engine_parameters = self.request.get("workflow_engine_parameters", {})

        # Obtain main workflow file to a path (no-scheme file URL).
        workflow_url = self.write_workflow(src_url=self.request["workflow_url"])

        if os.path.basename(workflow_url) == "workflow.zip" and zipfile.is_zipfile(workflow_url):
            # We've been sent a zip file. We should interpret this as an Amazon Genomics CLI-style zip file.
            # Extract everything next to the zip and find and open relvant files.
            logger.info("Extracting and parsing Amazon-style workflow bundle...")
            zip_info = amazon_wes_utils.parse_workflow_zip_file(workflow_url, self.wf_type)

            # Now parse Amazon's internal format into our own.

            # Find the real workflow source for the entrypoint file
            if 'workflowSource' in zip_info['files']:
                # They sent a file, which has been opened, so grab its path
                workflow_url = zip_info['files']['workflowSource'].name
                logger.info("Workflow source file: '%s'", workflow_url)
            elif 'workflowUrl' in zip_info['data']:
                # They are pointing to another URL.
                # TODO: What does Amazon expect this to mean? Are we supposed to recurse?
                # For now just forward it.
                workflow_url = zip_info['data']['workflowUrl']
                logger.info("Workflow reference URL: '%s'", workflow_url)
            else:
                # The parser is supposed to throw so we can't get here
                raise RuntimeError("Parser could not find workflow source or URL in zip")

            if 'workflowInputFiles' in zip_info['files'] and len(zip_info['files']['workflowInputFiles']) > 0:
                # The bundle contains a list of input files.
                # We interpret these as JSON, and layer them on top of each
                # other, and then apply workflow_params from the request.
                logger.info("Workflow came with %d bundled inputs files; coalescing final parameters",
                            len(zip_info['files']['workflowInputFiles']))

                coalesced_parameters = {}
                for binary_file in zip_info['files']['workflowInputFiles']:
                    try:
                        # Read each input file as a JSON
                        loaded_parameters = json.load(binary_file)
                    except json.JSONDecodeError as e:
                        raise RuntimeError(f"Could not parse inputs JSON {os.path.basename(binary_file.name)}: {e}")
                    # And merge them together in order
                    coalesced_parameters.update(loaded_parameters)
                # Then apply and replace the parameters that came with the request
                coalesced_parameters.update(workflow_params)
                workflow_params = coalesced_parameters

            if 'workflowOptions' in zip_info['files']:
                # The bundle contains an options JSON. We interpret these as
                # defaults for workflow_engine_parameters.
                logger.info(f"Workflow came with bundled options JSON")
                try:
                    # Read as a JSON
                    loaded_options = json.load(zip_info['files']['workflowOptions'])
                except json.JSONDecodeError as e:
                    raise RuntimeError(f"Could not parse options JSON: {e}")
                # Apply and replace the engine parameters that came with the
                # request.
                loaded_options.update(workflow_engine_parameters)
                workflow_engine_parameters = loaded_options

        # Write out CWL-style input file
        input_json = os.path.join(self.exec_dir, "wes_inputs.json")
        with open(input_json, "w") as f:
            json.dump(workflow_params, f)

        # create output directory
        if not os.path.exists(self.out_dir):
            os.makedirs(self.out_dir)

        # sort commands
        options = self.sort_options(workflow_engine_parameters)

        # construct and return the command to run
        if self.wf_type == "cwl":
            command_args = (
                ["toil-cwl-runner"] + options + [workflow_url, input_json]
            )
        elif self.wf_type == "wdl":
            command_args = (
                ["toil-wdl-runner"] + options + [workflow_url, input_json]
            )
        elif self.wf_type == "py":
            command_args = ["python", workflow_url] + options
        else:
            # This shouldn't happen if this was executed from the Toil WES server.
            raise RuntimeError(f"Unsupported workflow type: {self.wf_type}")

        return command_args

    def call_cmd(self, cmd: Union[List[str], str], cwd: str) -> "subprocess.Popen[bytes]":
        """
        Calls a command with Popen. Writes stdout, stderr, and the command to
        separate files.
        """
        stdout_f = os.path.join(self.scratch_dir, "stdout")
        stderr_f = os.path.join(self.scratch_dir, "stderr")

        with open(stdout_f, "w") as stdout, open(stderr_f, "w") as stderr:
            logger.info(f"Calling: '{' '.join(cmd)}'")
            process = subprocess.Popen(cmd, stdout=stdout, stderr=stderr, close_fds=True, cwd=cwd)

        return process

    def run(self) -> None:
        """
        Construct a command to run a the requested workflow with the options,
        run it, and deposit the outputs in the output directory.
        """

        # the task has been picked up by the runner and is currently preparing to run
        self.state_machine.send_initialize()
        commands = self.initialize_run()

        # store the job store location
        self.store.set("job_store", self.job_store)

        # Check if we are supposed to cancel
        state = self.get_state()
        if state in ("CANCELING", "CANCELED"):
            logger.info("Workflow canceled.")
            return

        # Otherwise start to run
        self.state_machine.send_run()
        process = self.call_cmd(cmd=commands, cwd=self.exec_dir)

        self.store.set("start_time", get_iso_time())
        self.store.set("cmd", " ".join(commands))

        try:
            exit_code = process.wait()
        except (KeyboardInterrupt, SystemExit, SoftTimeLimitExceeded) as e:
            logger.warning("Workflow interrupted: %s", type(e))
            process.terminate()

            try:
                process.wait(timeout=WAIT_FOR_DEATH_TIMEOUT)
            except subprocess.TimeoutExpired:
                # We need to actually stop now.
                process.kill()

            logger.info("Child process terminated by interruption.")
            exit_code = 130

        self.store.set("end_time", get_iso_time())
        self.store.set("exit_code", str(exit_code))

        logger.info('Toil child finished with code %s', exit_code)

        if exit_code == 0:
            self.state_machine.send_complete()
        # non-zero exit code indicates failure
        elif exit_code == 130:
            self.state_machine.send_canceled()
        else:
            self.state_machine.send_executor_error()

    def write_output_files(self) -> None:
        """
        Fetch all the files that this workflow generated and output information
        about them to `outputs.json`.
        """
        output_obj = {}
        job_store_type, _ = Toil.parseLocator(self.job_store)

        # For CWL workflows, the stdout should be a JSON object containing the outputs
        if self.wf_type == "cwl":
            try:
                with open(os.path.join(self.scratch_dir, "stdout")) as f:
                    output_obj = json.load(f)
            except Exception as e:
                logger.warning("Failed to read outputs object from stdout:", exc_info=e)

        elif job_store_type == "file":
            for file in os.listdir(self.out_dir):
                location = os.path.join(self.out_dir, file)
                output_obj[file] = {
                    "location": location,
                    "size": os.stat(location).st_size,
                    "class": get_file_class(location),
                }

        # TODO: fetch files from other job stores

        self.write_scratch_file("outputs.json", json.dumps(output_obj))

def run_wes_task(base_scratch_dir: str, state_store_url: str, workflow_id: str, request: Dict[str, Any], engine_options: List[str]) -> str:
    """
    Run a requested workflow.

    :param base_scratch_dir: Directory where the workflow's scratch dir will live, under the workflow's ID.

    :param state_store_url: URL/path at which the server and Celery task communicate about workflow state.

    :param workflow_id: ID of the workflow run.

    :returns: the state of the workflow run.
    """

    logger.info("Starting WES workflow")

    runner = ToilWorkflowRunner(base_scratch_dir, state_store_url, workflow_id,
                                request=request, engine_options=engine_options)

    try:
        runner.run()

        state = runner.get_state()
        logger.info(f"Workflow run completed with state: '{state}'.")

        if state == "COMPLETE":
            logger.info(f"Fetching output files.")
            runner.write_output_files()
    except (KeyboardInterrupt, SystemExit, SoftTimeLimitExceeded):
        # We canceled the workflow run
        logger.info('Canceling the workflow')
        runner.state_machine.send_canceled()
    except Exception:
        # The workflow run broke. We still count as the executor here.
        logger.exception('Running Toil produced an exception.')
        runner.state_machine.send_executor_error()
        raise

    return runner.get_state()

# Wrap the task function as a Celery task
run_wes = celery.task(name="run_wes")(run_wes_task)

def cancel_run(task_id: str) -> None:
    """
    Send a SIGTERM signal to the process that is running task_id.
    """
    celery.control.terminate(task_id, signal='SIGUSR1')

class TaskRunner:
    """
    Abstraction over the Celery API. Runs our run_wes task and allows canceling it.

    We can swap this out in the server to allow testing without Celery.
    """

    @staticmethod
    def run(args: Tuple[str, str, str, Dict[str, Any], List[str]], task_id: str) -> None:
        """
        Run the given task args with the given ID on Celery.
        """
        run_wes.apply_async(args=args,
                            task_id=task_id,
                            ignore_result=True)

    @staticmethod
    def cancel(task_id: str) -> None:
        """
        Cancel the task with the given ID on Celery.
        """
        cancel_run(task_id)

    @staticmethod
    def is_ok(task_id: str) -> bool:
        """
        Make sure that the task running system is working for the given task.
        If the task system has detected an internal failure, return False.
        """
        # Nothing to do for Celery
        return True

# If Celery can't be set up, we can just use this fake version instead.

class MultiprocessingTaskRunner(TaskRunner):
    """
    Version of TaskRunner that just runs tasks with Multiprocessing.

    Can't use threading because there's no way to send a cancel signal or
    exception to a Python thread, if loops in the task (i.e.
    ToilWorkflowRunner) don't poll for it.
    """

    _id_to_process: Dict[str, multiprocessing.Process] = {}
    _id_to_log: Dict[str, str] = {}

    @staticmethod
    def set_up_and_run_task(output_path: str, args: Tuple[str, str, str, Dict[str, Any], List[str]]) -> None:
        """
        Set up logging for the process into the given file and then call
        run_wes_task with the given arguments.

        If the process finishes successfully, it will clean up the log, but if
        the process crashes, the caller must clean up the log.
        """

        # Multiprocessing and the server manage to hide actual task output from
        # the tests. Logging messages will appear in pytest's "live" log but
        # not in the captured log. And replacing sys.stdout and sys.stderr
        # don't seem to work to collect tracebacks. So we make sure to
        # configure logging in the multiprocessing process and log to a file
        # that we were told about, so the server can come get the log if we
        # unexpectedly die.

        output_file = open(output_path, 'w')
        output_file.write('Initializing task log\n')
        output_file.flush()

        # Take over logging.
        root_logger = logging.getLogger()
        root_logger.setLevel(logging.DEBUG)
        handler = logging.StreamHandler(output_file)
        root_logger.addHandler(handler)

        def handle_sigterm(_: Any, __: Any) -> None:
            """
            Multiprocessing will send SIGTERM to stop us; translate that to
            something the run_wes_task shutdown code understands to avoid
            always waiting for the cancel timeout.

            Caller still needs to handle a sigterm exit code, in case we get
            canceled before the handler is set!
            """
            sys.exit()

        signal.signal(signal.SIGTERM, handle_sigterm)

        try:
            logger.info('Running task')
            output_file.flush()
            run_wes_task(*args)
        except Exception:
            logger.exception('Exception in task!')
            raise
        else:
            # If the task does not crash, clean up the log
            os.unlink(output_path)
        finally:
            logger.debug('Finishing task log')
            output_file.flush()
            output_file.close()

    @classmethod
    def run(cls, args: Tuple[str, str, str, Dict[str, Any], List[str]], task_id: str) -> None:
        """
        Run the given task args with the given ID.
        """

        # We need to send the child process's output somewhere or it will get lost during testing
        fd, path = tempfile.mkstemp()
        os.close(fd)
        # Store the log filename before the process, like is_ok() expects.
        cls._id_to_log[task_id] = path

        logger.info("Starting task %s in a process that should log to %s", task_id, path)

        cls._id_to_process[task_id] = multiprocessing.Process(target=cls.set_up_and_run_task, args=(path, args))
        cls._id_to_process[task_id].start()

    @classmethod
    def cancel(cls, task_id: str) -> None:
        """
        Cancel the task with the given ID.
        """
        if task_id in cls._id_to_process:
            logger.info("Stopping process for task %s", task_id)
            cls._id_to_process[task_id].terminate()
        else:
            logger.error("Tried to kill nonexistent task %s", task_id)

    @classmethod
    def is_ok(cls, task_id: str) -> bool:
        """
        Make sure that the task running system is working for the given task.
        If the task system has detected an internal failure, return False.
        """

        process = cls._id_to_process.get(task_id)
        if process is None:
            # Never heard of this task, so it's not broken.
            return True

        # If the process exited normally, or with an error code consistent with
        # being canceled by cancel(), then it is OK.
        ACCEPTABLE_EXIT_CODES = [0, -signal.SIGTERM]

        if process.exitcode is not None and process.exitcode not in ACCEPTABLE_EXIT_CODES:
            # Something went wring in the task and it couldn't handle it.
            logger.error("Process for running %s failed with code %s", task_id, process.exitcode)
            try:
                for line in open(cls._id_to_log[task_id]):
                    # Dump the task log
                    logger.error("Task said: %s", line.rstrip())
                # Remove the task log because it is no longer needed
                os.unlink(cls._id_to_log[task_id])
            except FileNotFoundError:
                # Log is already gone
                pass
            return False

        return True





# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import datetime
import errno
import filecmp
import inspect
import logging
import os
import random
import signal
import stat
import time
from abc import ABCMeta
from struct import pack, unpack
from uuid import uuid4

import pytest

from toil.common import Toil
from toil.fileStores import FileID
from toil.fileStores.cachingFileStore import (CacheUnbalancedError,
                                              IllegalDeletionCacheError)
from toil.job import Job
from toil.jobStores.abstractJobStore import NoSuchFileException
from toil.exceptions import FailedJobsException
from toil.realtimeLogger import RealtimeLogger
from toil.test import ToilTest, needs_aws_ec2, needs_google, slow

# Some tests take too long on the AWS jobstore and are unquitable for CI.  They can be
# be run during manual tests by setting this to False.
testingIsAutomatic = True

logger = logging.getLogger(__name__)


class hidden:
    """
    Hiding the abstract test classes from the Unittest loader so it can be inherited in different
    test suites for the different job stores.
    """
    class AbstractFileStoreTest(ToilTest, metaclass=ABCMeta):
        """
        An abstract base class for testing the various general functions described in
        :class:toil.fileStores.abstractFileStore.AbstractFileStore
        """
        # This is overwritten in the inheriting classs
        jobStoreType = None

        def _getTestJobStore(self):
            if self.jobStoreType == 'file':
                return self._getTestJobStorePath()
            elif self.jobStoreType == 'aws':
                return f'aws:{self.awsRegion()}:cache-tests-{str(uuid4())}'
            elif self.jobStoreType == 'google':
                projectID = os.getenv('TOIL_GOOGLE_PROJECTID')
                return f'google:{projectID}:cache-tests-{str(uuid4())}'
            else:
                raise RuntimeError('Illegal job store type.')

        def setUp(self):
            super().setUp()
            self.work_dir = self._createTempDir()
            self.options = Job.Runner.getDefaultOptions(self._getTestJobStore())
            self.options.logLevel = 'DEBUG'
            self.options.realTimeLogging = True
            self.options.workDir = self.work_dir
            self.options.clean = 'always'
            self.options.logFile = os.path.join(self.work_dir, 'logFile')

            self.tmp_dir = self._createTempDir()

        def create_file(self, content, executable=False):
            file_path = f'{self.tmp_dir}/{uuid4()}'

            with open(file_path, 'w') as f:
                f.write(content)

            if executable:
                # Add file owner execute permissions
                os.chmod(file_path, os.stat(file_path).st_mode | stat.S_IXUSR)

            return file_path

        @staticmethod
        def _uselessFunc(job):
            """
            I do nothing.  Don't judge me.
            """
            return None

        # Sanity test
        def testToilIsNotBroken(self):
            """
            Runs a simple DAG to test if if any features other that caching were broken.
            """
            A = Job.wrapJobFn(self._uselessFunc)
            B = Job.wrapJobFn(self._uselessFunc)
            C = Job.wrapJobFn(self._uselessFunc)
            D = Job.wrapJobFn(self._uselessFunc)
            A.addChild(B)
            A.addChild(C)
            B.addChild(D)
            C.addChild(D)
            Job.Runner.startToil(A, self.options)

        @slow
        def testFileStoreLogging(self):
            """
            Write a couple of files to the jobstore.  Delete a couple of them.  Read back written
            and locally deleted files.
            """

            class WatchingHandler(logging.Handler):
                """
                A logging handler that watches for a certain substring and
                trips a flag if it appears.
                """
                def __init__(self, match: str):
                    super().__init__()
                    self.match = match
                    self.seen = False

                def emit(self, record):
                    if self.match in record.getMessage():
                        self.seen = True

            handler = WatchingHandler("cats.txt")

            logging.getLogger().addHandler(handler)

            F = Job.wrapJobFn(self._accessAndFail,
                              disk='100M')
            try:
                Job.Runner.startToil(F, self.options)
            except FailedJobsException:
                # We expect this.
                pass

            logging.getLogger().removeHandler(handler)

            assert handler.seen, "Downloaded file name not found in logs of failing Toil run"

        @staticmethod
        def _accessAndFail(job):
            with job.fileStore.writeGlobalFileStream() as (writable, file_id):
                writable.write(b'Cats')
            localPath = os.path.join(job.fileStore.getLocalTempDir(), 'cats.txt')
            job.fileStore.readGlobalFile(file_id, localPath)
            with job.fileStore.readGlobalFileStream(file_id) as readable:
                pass
            raise RuntimeError("I do not like this file")


        # Test filestore operations.  This is a slightly less intense version of the cache specific
        # test `testReturnFileSizes`
        @slow
        def testFileStoreOperations(self):
            """
            Write a couple of files to the jobstore.  Delete a couple of them.  Read back written
            and locally deleted files.
            """
            workdir = self._createTempDir(purpose='nonLocalDir')
            F = Job.wrapJobFn(self._testFileStoreOperations,
                              nonLocalDir=workdir,
                              numIters=30, disk='2G')
            Job.Runner.startToil(F, self.options)

        @staticmethod
        def _testFileStoreOperations(job, nonLocalDir, numIters=100):
            """
            Aux function for testFileStoreOperations Conduct numIters operations.
            """
            work_dir = job.fileStore.getLocalTempDir()
            writtenFiles = {}  # fsID: (size, isLocal)
            localFileIDs = set()
            # Add one file for the sake of having something in the job store
            writeFileSize = random.randint(0, 30)
            cls = hidden.AbstractNonCachingFileStoreTest
            fsId, _ = cls._writeFileToJobStore(job, isLocalFile=True, nonLocalDir=nonLocalDir,
                                               fileMB=writeFileSize)

            # Fill in the size of the local file we just made
            writtenFiles[fsId] = writeFileSize
            # Remember it actually should be local
            localFileIDs.add(fsId)
            logger.info('Now have local file: %s', fsId)

            i = 0
            while i <= numIters:
                randVal = random.random()
                if randVal < 0.33:  # Write
                    writeFileSize = random.randint(0, 30)
                    isLocalFile = True if random.random() <= 0.5 else False
                    fsID, _ = cls._writeFileToJobStore(job, isLocalFile=isLocalFile,
                                                       nonLocalDir=nonLocalDir,
                                                       fileMB=writeFileSize)
                    writtenFiles[fsID] = writeFileSize
                    if isLocalFile:
                        localFileIDs.add(fsID)
                    logger.info('Wrote %s file of size %d MB: %s', 'local' if isLocalFile else 'non-local', writeFileSize, fsID)
                else:
                    if len(writtenFiles) == 0:
                        continue
                    else:
                        fsID, rdelFileSize = random.choice(list(writtenFiles.items()))
                        rdelRandVal = random.random()
                    if randVal < 0.66:  # Read
                        mutable = True if random.random() <= 0.5 else False
                        cache = True if random.random() <= 0.5 else False
                        job.fileStore.readGlobalFile(fsID, '/'.join([work_dir, str(uuid4())]),
                                                     cache=cache, mutable=mutable)
                        localFileIDs.add(fsID)
                        logger.info('Read %s %s local copy of: %s', 'mutable' if mutable else 'immutable', 'cached' if cache else 'uncached', fsID)
                    else:  # Delete
                        if rdelRandVal <= 0.5:  # Local Delete
                            if fsID not in localFileIDs:
                                # Make sure trying to deleteLocalFile this file fails properly
                                try:
                                    job.fileStore.deleteLocalFile(fsID)
                                except OSError as e:
                                    if e.errno != errno.ENOENT:
                                        # This is supposed to produce an
                                        # ENOENT. If it doesn't something is
                                        # broken.
                                        raise
                                    logger.info('Correctly fail to local-delete non-local file: %s', fsID)
                                else:
                                    assert False, f"Was able to delete non-local file {fsID}"
                            else:
                                logger.info('Delete local file: %s', fsID)
                                job.fileStore.deleteLocalFile(fsID)
                        else:  # Global Delete
                            job.fileStore.deleteGlobalFile(fsID)
                            writtenFiles.pop(fsID)
                        if fsID in localFileIDs:
                            localFileIDs.remove(fsID)
                            logger.info('No longer have file: %s', fsID)
                i += 1

        def testWriteReadGlobalFilePermissions(self):
            """
            Ensures that uploaded files preserve their file permissions when they
            are downloaded again. This function checks that a written executable file
            maintains its executability after being read.
            """
            for executable in True, False:
                for caching in True, False:
                    with self.subTest(f'Testing readwrite file permissions\n'
                                      f'[executable: {executable}]\n'
                                      f'[caching: {caching}]\n'):
                        self.options.caching = caching
                        read_write_job = Job.wrapJobFn(self._testWriteReadGlobalFilePermissions, executable=executable)
                        Job.Runner.startToil(read_write_job, self.options)

        @staticmethod
        def _testWriteReadGlobalFilePermissions(job, executable):
            srcFile = job.fileStore.getLocalTempFile()
            with open(srcFile, 'w') as f:
                f.write('Hello')

            if executable:
                os.chmod(srcFile, os.stat(srcFile).st_mode | stat.S_IXUSR)

            # Initial file owner execute permissions
            initialPermissions = os.stat(srcFile).st_mode & stat.S_IXUSR
            fileID = job.fileStore.writeGlobalFile(srcFile)

            for mutable in True, False:
                for symlink in True, False:
                    dstFile = job.fileStore.getLocalTempFileName()
                    job.fileStore.readGlobalFile(fileID, userPath=dstFile, mutable=mutable, symlink=symlink)
                    # Current file owner execute permissions
                    currentPermissions = os.stat(dstFile).st_mode & stat.S_IXUSR
                    assert initialPermissions == currentPermissions, f'{initialPermissions} != {currentPermissions}'

        def testWriteExportFileCompatibility(self):
            """
            Ensures that files created in a job preserve their executable permissions
            when they are exported from the leader.
            """
            for executable in True, False:
                export_file_job = Job.wrapJobFn(self._testWriteExportFileCompatibility, executable=executable)
                with Toil(self.options) as toil:
                    initialPermissions, fileID = toil.start(export_file_job)
                    dstFile = os.path.join(self._createTempDir(), str(uuid4()))
                    toil.exportFile(fileID, 'file://' + dstFile)
                    currentPermissions = os.stat(dstFile).st_mode & stat.S_IXUSR

                    assert initialPermissions == currentPermissions, f'{initialPermissions} != {currentPermissions}'

        @staticmethod
        def _testWriteExportFileCompatibility(job, executable):
            srcFile = job.fileStore.getLocalTempFile()
            with open(srcFile, 'w') as f:
                f.write('Hello')
            if executable:
                os.chmod(srcFile, os.stat(srcFile).st_mode | stat.S_IXUSR)
            initialPermissions = os.stat(srcFile).st_mode & stat.S_IXUSR
            fileID = job.fileStore.writeGlobalFile(srcFile)
            return initialPermissions, fileID

        def testImportReadFileCompatibility(self):
            """
            Ensures that files imported to the leader preserve their executable permissions
            when they are read by the fileStore.
            """
            with Toil(self.options) as toil:
                for executable in True, False:
                    file_path = self.create_file(content='Hello', executable=executable)
                    initial_permissions = os.stat(file_path).st_mode & stat.S_IXUSR
                    file_id = toil.importFile(f'file://{file_path}')
                    for mutable in True, False:
                        for symlink in True, False:
                            with self.subTest(f'Now testing readGlobalFileWith: mutable={mutable} symlink={symlink}'):
                                A = Job.wrapJobFn(self._testImportReadFileCompatibility,
                                                  fileID=file_id,
                                                  initialPermissions=initial_permissions,
                                                  mutable=mutable,
                                                  symlink=symlink)
                                toil.start(A)

        @staticmethod
        def _testImportReadFileCompatibility(job, fileID, initialPermissions, mutable, symlink):
            dstFile = job.fileStore.readGlobalFile(fileID, mutable=mutable, symlink=symlink)
            currentPermissions = os.stat(dstFile).st_mode & stat.S_IXUSR

            assert initialPermissions == currentPermissions

        def testReadWriteFileStreamTextMode(self):
            """
            Checks if text mode is compatible with file streams.
            """
            with Toil(self.options) as toil:
                A = Job.wrapJobFn(self._testReadWriteFileStreamTextMode)
                toil.start(A)

        @staticmethod
        def _testReadWriteFileStreamTextMode(job):
            with job.fileStore.writeGlobalFileStream(encoding='utf-8') as (stream, fileID):
                stream.write('foo')
            job.fileStore.readGlobalFileStream(fileID)
            with job.fileStore.readGlobalFileStream(fileID, encoding='utf-8') as stream2:
                assert 'foo' == stream2.read()

        @staticmethod
        def _writeFileToJobStore(job, isLocalFile, nonLocalDir=None, fileMB=1):
            """
            This function creates a file and writes it to the jobstore.

            :param bool isLocalFile: Is the file local(T) or Non-Local(F)?
            :param str nonLocalDir: A dir to write the file to.  If unspecified, a local directory
                                    is created.
            :param int fileMB: Size of the created file in MB
            """
            if isLocalFile:
                work_dir = job.fileStore.getLocalTempDir()
            else:
                assert nonLocalDir is not None
                work_dir = nonLocalDir
            with open(os.path.join(work_dir, str(uuid4())), 'wb') as testFile:
                testFile.write(os.urandom(fileMB * 1024 * 1024))

            return job.fileStore.writeGlobalFile(testFile.name), testFile

    class AbstractNonCachingFileStoreTest(AbstractFileStoreTest, metaclass=ABCMeta):
        """
        Abstract tests for the the various functions in
        :class:toil.fileStores.nonCachingFileStore.NonCachingFileStore. These
        tests are general enough that they can also be used for
        :class:toil.fileStores.CachingFileStore.
        """

        def setUp(self):
            super().setUp()
            self.options.caching = False

    class AbstractCachingFileStoreTest(AbstractFileStoreTest, metaclass=ABCMeta):
        """
        Abstract tests for the the various cache-related functions in
        :class:toil.fileStores.cachingFileStore.CachingFileStore.
        """

        def setUp(self):
            super().setUp()
            self.options.caching = True

        @slow
        def testExtremeCacheSetup(self):
            """
            Try to create the cache with bad worker active and then have 10 child jobs try to run in
            the chain.  This tests whether the cache is created properly even when the job crashes
            randomly.
            """
            if testingIsAutomatic and self.jobStoreType != 'file':
                self.skipTest("To save time")
            self.options.retryCount = 10
            self.options.badWorker = 0.25
            self.options.badWorkerFailInterval = 0.2
            for test in range(0, 20):
                E = Job.wrapJobFn(self._uselessFunc)
                F = Job.wrapJobFn(self._uselessFunc)
                jobs = {}
                for i in range(0, 10):
                    jobs[i] = Job.wrapJobFn(self._uselessFunc)
                    E.addChild(jobs[i])
                    jobs[i].addChild(F)
                Job.Runner.startToil(E, self.options)

        @slow
        def testCacheEvictionPartialEvict(self):
            """
            Ensure the cache eviction happens as expected.  Two files (20MB and 30MB) are written
            sequentially into the job store in separate jobs.  The cache max is force set to 50MB.
            A Third Job requests 10MB of disk requiring eviction of the 1st file.  Ensure that the
            behavior is as expected.
            """
            self._testValidityOfCacheEvictTest()

            # Explicitly set clean to always so even the failed cases get cleaned (This will
            # overwrite the value set in setUp if it is ever changed in the future)
            self.options.clean = 'always'

            self._testCacheEviction(file1MB=20, file2MB=30, diskRequestMB=10)

        @slow
        def testCacheEvictionTotalEvict(self):
            """
            Ensure the cache eviction happens as expected.  Two files (20MB and 30MB) are written
            sequentially into the job store in separate jobs.  The cache max is force set to 50MB.
            A Third Job requests 10MB of disk requiring eviction of the 1st file.  Ensure that the
            behavior is as expected.
            """
            self._testValidityOfCacheEvictTest()

            # Explicitly set clean to always so even the failed cases get cleaned (This will
            # overwrite the value set in setUp if it is ever changed in the future)
            self.options.clean = 'always'

            self._testCacheEviction(file1MB=20, file2MB=30, diskRequestMB=30)

        @slow
        def testCacheEvictionFailCase(self):
            """
            Ensure the cache eviction happens as expected.  Two files (20MB and 30MB) are written
            sequentially into the job store in separate jobs.  The cache max is force set to 50MB.
            A Third Job requests 10MB of disk requiring eviction of the 1st file.  Ensure that the
            behavior is as expected.
            """
            self._testValidityOfCacheEvictTest()

            # Explicitly set clean to always so even the failed cases get cleaned (This will
            # overwrite the value set in setUp if it is ever changed in the future)
            self.options.clean = 'always'

            self._testCacheEviction(file1MB=20, file2MB=30, diskRequestMB=60)

        def _testValidityOfCacheEvictTest(self):
            # If the job store and cache are on the same file system, file
            # sizes are accounted for by the job store and are not reflected in
            # the cache hence this test is redundant (caching will be free).
            if not self.options.jobStore.startswith(('aws', 'google')):
                workDirDev = os.stat(self.options.workDir).st_dev
                jobStoreDev = os.stat(os.path.dirname(self.options.jobStore)).st_dev
                if workDirDev == jobStoreDev:
                    self.skipTest('Job store and working directory are on the same filesystem.')

        def _testCacheEviction(self, file1MB, file2MB, diskRequestMB):
            """
            Ensure the cache eviction happens as expected.  Two files (20MB and 30MB) are written
            sequentially into the job store in separate jobs.  The cache max is force set to 50MB.
            A Third Job requests either 10, 30 or 60MB -- requiring eviction of 1 file, both files,
            or results in an error due to lack of space, respectively.  Ensure that the behavior is
            as expected.
            """
            self.options.retryCount = 0
            if diskRequestMB > 50:
                # This can be non int as it will never reach _probeJobReqs
                expectedResult = 'Fail'
            else:
                expectedResult = 50 - file1MB if diskRequestMB <= file1MB else 0
            try:
                A = Job.wrapJobFn(self._writeFileToJobStoreWithAsserts, isLocalFile=True,
                                  fileMB=file1MB)
                # Sleep for 1 second after writing the first file so that their ctimes are
                # guaranteed to be distinct for the purpose of this test.
                B = Job.wrapJobFn(self._sleepy, timeToSleep=1)
                C = Job.wrapJobFn(self._writeFileToJobStoreWithAsserts, isLocalFile=True,
                                  fileMB=file2MB)
                D = Job.wrapJobFn(self._adjustCacheLimit, newTotalMB=50, disk='0Mi')
                E = Job.wrapJobFn(self._uselessFunc, disk=''.join([str(diskRequestMB), 'Mi']))
                # Set it to > 2GB such that the cleanup jobs don't die in the non-fail cases
                F = Job.wrapJobFn(self._adjustCacheLimit, newTotalMB=5000, disk='10Mi')
                G = Job.wrapJobFn(self._probeJobReqs, sigmaJob=100, cached=expectedResult,
                                  disk='100Mi')
                A.addChild(B)
                B.addChild(C)
                C.addChild(D)
                D.addChild(E)
                E.addChild(F)
                F.addChild(G)
                Job.Runner.startToil(A, self.options)
            except FailedJobsException as err:
                with open(self.options.logFile) as f:
                    logContents = f.read()
                if CacheUnbalancedError.message in logContents:
                    self.assertEqual(expectedResult, 'Fail')
                else:
                    self.fail('Toil did not raise the expected CacheUnbalancedError but failed for some other reason')

        @staticmethod
        def _writeFileToJobStoreWithAsserts(job, isLocalFile, nonLocalDir=None, fileMB=1, expectAsyncUpload=True):
            """
            This function creates a file and writes it to the jobstore.

            :param bool isLocalFile: Is the file local(T) (i.e. in the file
                                     store managed temp dir) or Non-Local(F)?
                                     Non-local files should not be cached.
            :param str nonLocalDir: A dir to write the file to.  If unspecified, a local directory
                                    is created.
            :param int fileMB: Size of the created file in MB
            :param bool expectAsyncUpload: Whether we expect the upload to hit
                                           the job store later(T) or immediately(F)
            """
            cls = hidden.AbstractNonCachingFileStoreTest
            fsID, testFile = cls._writeFileToJobStore(job, isLocalFile, nonLocalDir, fileMB)
            actual = os.stat(testFile.name).st_nlink

            # If the caching is free, the job store must have hard links to
            # everything the file store has.
            expectJobStoreLink = job.fileStore.cachingIsFree()

            # How many links ought this file to have?
            expected = 1

            if isLocalFile:
                # We expect a hard link into the cache and not a copy
                expected += 1

                if expectJobStoreLink and not expectAsyncUpload:
                    # We also expect a link in the job store
                    expected += 1

            assert actual == expected, 'Should have %d links. Got %d.' % (expected, actual)

            logger.info('Uploaded %s with %d links', fsID, actual)

            if not isLocalFile:
                # Make sure it isn't cached if we don't want it to be
                assert not job.fileStore.fileIsCached(fsID), "File uploaded from non-local-temp directory %s should not be cached" % nonLocalDir

            return fsID

        @staticmethod
        def _sleepy(job, timeToSleep):
            """
            I'm waiting for prince charming... but only for timeToSleep seconds.

            :param int timeToSleep: Time in seconds
            """
            time.sleep(timeToSleep)

        @staticmethod
        def _adjustCacheLimit(job, newTotalMB):
            """
            This function tells the cache to adopt a new "total" value =
            newTotalMB, changing the maximum cache disk space allowed for the
            run.

            :param int newTotalMB: New total cache disk space limit in MB.
            """

            # Convert to bytes and pass on to the actual cache
            job.fileStore.adjustCacheLimit(float(newTotalMB * 1024 * 1024))

        @staticmethod
        def _probeJobReqs(job, total=None, cached=None, sigmaJob=None):
            """
            Probes the cacheLockFile to ensure the values for total, disk and cache are as expected.
            Can also specify combinations of the requirements if desired.

            :param int total: Expected Total Space available for caching in MB.
            :param int cached: Expected Total size of files in the cache in MB.
            :param int sigmaJob: Expected sum of job requirements in MB.
            """

            RealtimeLogger.info('Probing job requirements')

            valueDict = locals()
            assert (total or cached or sigmaJob)

            # Work out which function to call for which value
            toCall = {'total': job.fileStore.getCacheLimit,
                      'cached': job.fileStore.getCacheUsed,
                      'sigmaJob': job.fileStore.getCacheExtraJobSpace}

            for value in ('total', 'cached', 'sigmaJob'):
                # If the value wasn't provided, it is None and should be ignored
                if valueDict[value] is None:
                    continue

                RealtimeLogger.info('Probing cache state: %s', value)

                expectedBytes = valueDict[value] * 1024 * 1024
                cacheInfoBytes = toCall[value]()

                RealtimeLogger.info('Got %d for %s; expected %d', cacheInfoBytes, value, expectedBytes)

                assert cacheInfoBytes == expectedBytes, 'Testing %s: Expected ' % value + \
                                                  f'{expectedBytes} but got {cacheInfoBytes}.'

        @slow
        def testAsyncWriteWithCaching(self):
            """
            Ensure the Async Writing of files happens as expected.  The first Job forcefully
            modifies the cache size to 1GB. The second asks for 1GB of disk and  writes a 900MB
            file into cache then rewrites it to the job store triggering an async write since the
            two unique jobstore IDs point to the same local file.  Also, the second write is not
            cached since the first was written to cache, and there "isn't enough space" to cache the
            second.  Imediately assert that the second write isn't cached, and is being
            asynchronously written to the job store.

            Attempting to get the file from the jobstore should not fail.
            """
            self.options.retryCount = 0
            self.options.logLevel = 'DEBUG'
            A = Job.wrapJobFn(self._adjustCacheLimit, newTotalMB=1024, disk='1G')
            B = Job.wrapJobFn(self._doubleWriteFileToJobStore, fileMB=850, disk='900M')
            C = Job.wrapJobFn(self._readFromJobStoreWithoutAssertions, fsID=B.rv(), disk='1G')
            # Set it to > 2GB such that the cleanup jobs don't die.
            D = Job.wrapJobFn(self._adjustCacheLimit, newTotalMB=5000, disk='1G')
            A.addChild(B)
            B.addChild(C)
            C.addChild(D)
            Job.Runner.startToil(A, self.options)

        @staticmethod
        def _doubleWriteFileToJobStore(job, fileMB):
            """
            Write a local file to job store, then write it again.  The second should trigger an
            async write.

            :param job: job
            :param fileMB: File Size
            :return: Job store file ID for second written file
            """
            job.fileStore.logToMaster('Double writing a file into job store')
            work_dir = job.fileStore.getLocalTempDir()
            with open(os.path.join(work_dir, str(uuid4())), 'wb') as testFile:
                testFile.write(os.urandom(fileMB * 1024 * 1024))

            job.fileStore.logToMaster('Writing copy 1 and discarding ID')
            job.fileStore.writeGlobalFile(testFile.name)
            job.fileStore.logToMaster('Writing copy 2 and saving ID')
            fsID = job.fileStore.writeGlobalFile(testFile.name)
            job.fileStore.logToMaster(f'Copy 2 ID: {fsID}')

            hidden.AbstractCachingFileStoreTest._readFromJobStoreWithoutAssertions(job, fsID)

            job.fileStore.logToMaster('Writing copy 3 and returning ID')
            return job.fileStore.writeGlobalFile(testFile.name)

        @staticmethod
        def _readFromJobStoreWithoutAssertions(job, fsID):
            """
            Reads a file from the job store.  That will be all, thank you.

            :param job: job
            :param fsID: Job store file ID for the read file
            :return: None
            """
            job.fileStore.logToMaster('Reading the written file')
            job.fileStore.readGlobalFile(fsID)

        # writeGlobalFile tests

        def testWriteNonLocalFileToJobStore(self):
            """
            Write a file not in localTempDir to the job store.  Such a file should not be cached.
            Ensure the file is not cached.
            """
            workdir = self._createTempDir(purpose='nonLocalDir')
            A = Job.wrapJobFn(self._writeFileToJobStoreWithAsserts, isLocalFile=False,
                              nonLocalDir=workdir)
            Job.Runner.startToil(A, self.options)

        def testWriteLocalFileToJobStore(self):
            """
            Write a file from the localTempDir to the job store.  Such a file will be cached by
            default.  Ensure the file is cached.
            """
            A = Job.wrapJobFn(self._writeFileToJobStoreWithAsserts, isLocalFile=True)
            Job.Runner.startToil(A, self.options)

        # readGlobalFile tests

        def testReadCacheMissFileFromJobStoreWithoutCachingReadFile(self):
            """
            Read a file from the file store that does not have a corresponding cached copy.  Do not
            cache the read file.  Ensure the number of links on the file are appropriate.
            """
            self._testCacheMissFunction(cacheReadFile=False)

        def testReadCacheMissFileFromJobStoreWithCachingReadFile(self):
            """
            Read a file from the file store that does not have a corresponding cached copy.  Cache
            the read file.  Ensure the number of links on the file are appropriate.
            """
            self._testCacheMissFunction(cacheReadFile=True)

        def _testCacheMissFunction(self, cacheReadFile):
            """
            This is the function that actually does what the 2 cache miss functions want.

            :param cacheReadFile: Does the read file need to be cached(T) or not(F)
            """
            workdir = self._createTempDir(purpose='nonLocalDir')
            A = Job.wrapJobFn(self._writeFileToJobStoreWithAsserts, isLocalFile=False,
                              nonLocalDir=workdir)
            B = Job.wrapJobFn(self._readFromJobStore, isCachedFile=False,
                              cacheReadFile=cacheReadFile, fsID=A.rv())
            A.addChild(B)
            Job.Runner.startToil(A, self.options)

        @staticmethod
        def _readFromJobStore(job, isCachedFile, cacheReadFile, fsID, isTest=True):
            """
            Read a file from the filestore.  If the file was cached, ensure it was hard linked
            correctly.  If it wasn't, ensure it was put into cache.

            Note that we may see hard links when we don't expect it based on
            caching, because immutable reads from the FileJobStore can be
            fulfilled by hardlinks. We only do immutable reads.

            :param bool isCachedFile: Flag.  Was the read file read from cache(T)? If so, we look for a hard link.
            :param bool cacheReadFile: Should the the file that is read be cached(T)?
            :param str fsID: job store file ID
            :param bool isTest: Is this being run as a test(T) or an accessory to another test(F)?

            """
            work_dir = job.fileStore.getLocalTempDir()
            wantHardLink = False
            if isCachedFile:
                outfile = job.fileStore.readGlobalFile(fsID, '/'.join([work_dir, 'temp']),
                                                       mutable=False)
                wantHardLink = True
            else:
                if cacheReadFile:
                    outfile = job.fileStore.readGlobalFile(fsID, '/'.join([work_dir, 'temp']),
                                                           cache=True, mutable=False)
                    wantHardLink = True
                else:
                    assert not job.fileStore.fileIsCached(fsID), "File mistakenly cached before read"
                    outfile = job.fileStore.readGlobalFile(fsID, '/'.join([work_dir, 'temp']),
                                                           cache=False, mutable=False)
                    assert not job.fileStore.fileIsCached(fsID), "File mistakenly cached after read"
                    wantHardLink = False
            if isTest:
                actual = os.stat(outfile).st_nlink
                if wantHardLink:
                    assert actual > 1, 'Should have multiple links for file that was %s and %s. Got %i.' % ('cached' if isCachedFile else 'not cached',
                        'saved' if cacheReadFile else 'not saved', actual)
                # We need to accept harf links even if we don't want them,
                # because we may get them straight from the FileJobStore since
                # we asked for immutable reads.
                return None
            else:
                return outfile

        def testReadCachHitFileFromJobStore(self):
            """
            Read a file from the file store that has a corresponding cached copy.  Ensure the number
            of links on the file are appropriate.
            """
            A = Job.wrapJobFn(self._writeFileToJobStoreWithAsserts, isLocalFile=True)
            B = Job.wrapJobFn(self._readFromJobStore, isCachedFile=True, cacheReadFile=None,
                              fsID=A.rv())
            A.addChild(B)
            Job.Runner.startToil(A, self.options)

        @slow
        def testMultipleJobsReadSameCacheHitGlobalFile(self):
            """
            Write a local file to the job store (hence adding a copy to cache), then have 10 jobs
            read it. Assert cached file size never goes up, assert unused job
            required disk space is always:
                   (a multiple of job reqs) - (number of current file readers * filesize).
            At the end, assert the cache shows unused job-required space = 0.
            """
            self._testMultipleJobsReadGlobalFileFunction(cacheHit=True)

        @slow
        def testMultipleJobsReadSameCacheMissGlobalFile(self):
            """
            Write a non-local file to the job store(hence no cached copy), then have 10 jobs read
            it. Assert cached file size never goes up, assert unused job
            required disk space is always:
                   (a multiple of job reqs) - (number of current file readers * filesize).
            At the end, assert the cache shows unused job-required space = 0.
            """
            self._testMultipleJobsReadGlobalFileFunction(cacheHit=False)

        def _testMultipleJobsReadGlobalFileFunction(self, cacheHit):
            """
            This function does what the two Multiple File reading tests want to do

            :param bool cacheHit: Is the test for the CacheHit case(T) or cacheMiss case(F)
            """
            dirPurpose = 'tempWriteDir' if cacheHit else 'nonLocalDir'
            workdir = self._createTempDir(purpose=dirPurpose)
            with open(os.path.join(workdir, 'test'), 'w') as x:
                x.write(str(0))
            A = Job.wrapJobFn(self._writeFileToJobStoreWithAsserts, isLocalFile=cacheHit,
                              nonLocalDir=workdir,
                              fileMB=256)
            B = Job.wrapJobFn(self._probeJobReqs, sigmaJob=100, disk='100Mi')
            jobs = {}
            for i in range(0, 10):
                jobs[i] = Job.wrapJobFn(self._multipleFileReader, diskMB=1024, fsID=A.rv(),
                                        maxWriteFile=os.path.abspath(x.name), disk='1Gi',
                                        memory='10Mi', cores=1)
                A.addChild(jobs[i])
                jobs[i].addChild(B)
            Job.Runner.startToil(A, self.options)
            with open(x.name) as y:
                assert int(y.read()) > 2

        @staticmethod
        def _multipleFileReader(job, diskMB, fsID, maxWriteFile):
            """
            Read a file from the job store immutable and explicitly ask to have it in the cache.
            If caching files is free, assert used cache space is zero, else
            assert it is equal to the read file.
            Also assert the sum job reqs + (number of readers of file * filesize) is
            and integer multiple of the disk requirements provided to this job.

            :param int diskMB: disk requirements provided to the job
            :param str fsID: job store file ID
            :param str maxWriteFile: path to file where the max number of concurrent readers of
                                     file will be written
            """
            work_dir = job.fileStore.getLocalTempDir()
            outfile = job.fileStore.readGlobalFile(fsID, '/'.join([work_dir, 'temp']), cache=True,
                                                   mutable=False)
            diskBytes = diskMB * 1024 * 1024
            fileStats = os.stat(outfile)
            fileSize = fileStats.st_size

            currentReaders = job.fileStore.getFileReaderCount(fsID)

            extraJobSpace = job.fileStore.getCacheExtraJobSpace()

            usedCache = job.fileStore.getCacheUsed()

            logger.info('Extra job space: %s', str(extraJobSpace))
            logger.info('Current file readers: %s', str(currentReaders))
            logger.info('File size: %s', str(fileSize))
            logger.info('Job disk bytes: %s', str(diskBytes))
            logger.info('Used cache: %s', str(usedCache))

            with open(maxWriteFile, 'r+') as x:
                prev_max = int(x.read())
                x.seek(0)
                x.truncate()
                x.write(str(max(prev_max, currentReaders)))
            if job.fileStore.cachingIsFree():
                # No space should be used when caching is free
                assert usedCache == 0.0
            else:
                # The right amount of space should be used otherwise
                assert usedCache == fileSize

            # Make sure that there's no over-usage of job requirements
            assert ((extraJobSpace + currentReaders * fileSize) %
                    diskBytes) == 0.0
            # Sleep so there's no race conditions where a job ends before another can get a hold of
            # the file
            time.sleep(3)

        @staticmethod
        def _writeExportGlobalFile(job):
            fileName = os.path.join(job.fileStore.getLocalTempDir(), 'testfile')
            with open(fileName, 'wb') as f:
                f.write(os.urandom(1024 * 30000)) # 30 Mb
            outputFile = os.path.join(job.fileStore.getLocalTempDir(), 'exportedFile')
            job.fileStore.export_file(job.fileStore.writeGlobalFile(fileName), 'File://' + outputFile)
            if not filecmp.cmp(fileName, outputFile):
                logger.warning('Source file: %s', str(os.stat(fileName)))
                logger.warning('Destination file: %s', str(os.stat(outputFile)))
                raise RuntimeError(f"File {fileName} did not properly get copied to {outputFile}")

        @slow
        def testFileStoreExportFile(self):
            # Tests that files written to job store can be immediately exported
            # motivated by https://github.com/BD2KGenomics/toil/issues/1469
            root = Job.wrapJobFn(self._writeExportGlobalFile)
            Job.Runner.startToil(root, self.options)

        # Testing for the return of file sizes to the sigma job pool.
        @slow
        def testReturnFileSizes(self):
            """
            Write a couple of files to the jobstore. Delete a couple of them.
            Read back written and locally deleted files. Ensure that after
            every step that the cache is in a valid state.
            """
            workdir = self._createTempDir(purpose='nonLocalDir')
            F = Job.wrapJobFn(self._returnFileTestFn,
                              jobDisk=2 * 1024 * 1024 * 1024,
                              initialCachedSize=0,
                              nonLocalDir=workdir,
                              disk='2Gi')
            Job.Runner.startToil(F, self.options)

        @slow
        def testReturnFileSizesWithBadWorker(self):
            """
            Write a couple of files to the jobstore. Delete a couple of them.
            Read back written and locally deleted files. Ensure that after
            every step that the cache is in a valid state.
            """
            self.options.retryCount = 20
            self.options.badWorker = 0.5
            self.options.badWorkerFailInterval = 0.1
            workdir = self._createTempDir(purpose='nonLocalDir')
            F = Job.wrapJobFn(self._returnFileTestFn,
                              jobDisk=2 * 1024 * 1024 * 1024,
                              initialCachedSize=0,
                              nonLocalDir=workdir,
                              numIters=30, disk='2Gi')
            Job.Runner.startToil(F, self.options)

        @staticmethod
        def _returnFileTestFn(job, jobDisk, initialCachedSize, nonLocalDir, numIters=100):
            """
            Aux function for jobCacheTest.testReturnFileSizes Conduct numIters operations and ensure
            the cache has the right amount of data in it at all times.

            Track the cache calculations even thought they won't be used in filejobstore

            Assumes nothing is evicted from the cache.

            :param float jobDisk: The value of disk passed to this job.
            """
            cached = initialCachedSize
            RealtimeLogger.info('Expecting %d bytes cached initially', cached)
            work_dir = job.fileStore.getLocalTempDir()
            writtenFiles = {}  # fsID: (size, isLocal)
            # fsid: local/mutable/immutable for all operations that should make local files as tracked by the FileStore
            localFileIDs = collections.defaultdict(list)
            # Add one file for the sake of having something in the job store
            writeFileSize = random.randint(0, 30)
            jobDisk -= writeFileSize * 1024 * 1024
            # We keep jobDisk in sync with the amount of free space the job
            # still has that the file store doesn't know it has used.
            cls = hidden.AbstractCachingFileStoreTest
            fsId = cls._writeFileToJobStoreWithAsserts(job, isLocalFile=True, fileMB=writeFileSize)
            writtenFiles[fsId] = writeFileSize
            if job.fileStore.fileIsCached(list(writtenFiles.keys())[0]):
                cached += writeFileSize * 1024 * 1024
                RealtimeLogger.info('Expecting %d bytes cached because file of %d MB is cached', cached, writeFileSize)
            else:
                RealtimeLogger.info('Expecting %d bytes cached because file of %d MB is not cached', cached, writeFileSize)
            localFileIDs[list(writtenFiles.keys())[0]].append('local')
            RealtimeLogger.info('Checking for %d bytes cached', cached)
            cls._requirementsConcur(job, jobDisk, cached)
            i = 0
            while i <= numIters:
                randVal = random.random()
                if randVal < 0.33:  # Write
                    RealtimeLogger.info('Writing a file')
                    writeFileSize = random.randint(0, 30)
                    if random.random() <= 0.5:  # Write a local file
                        RealtimeLogger.info('Writing a local file of %d MB', writeFileSize)
                        fsID = cls._writeFileToJobStoreWithAsserts(job, isLocalFile=True,
                                                                   fileMB=writeFileSize)
                        RealtimeLogger.info('Wrote local file: %s', fsID)
                        writtenFiles[fsID] = writeFileSize
                        localFileIDs[fsID].append('local')
                        jobDisk -= writeFileSize * 1024 * 1024
                        if job.fileStore.fileIsCached(fsID):
                            cached += writeFileSize * 1024 * 1024
                            RealtimeLogger.info('Expecting %d bytes cached because file of %d MB is cached', cached, writeFileSize)
                        else:
                            RealtimeLogger.info('Expecting %d bytes cached because file of %d MB is not cached', cached, writeFileSize)
                    else:  # Write a non-local file
                        RealtimeLogger.info('Writing a non-local file of %d MB', writeFileSize)
                        fsID = cls._writeFileToJobStoreWithAsserts(job, isLocalFile=False,
                                                                   nonLocalDir=nonLocalDir,
                                                                   fileMB=writeFileSize)
                        RealtimeLogger.info('Wrote non-local file: %s', fsID)
                        writtenFiles[fsID] = writeFileSize
                        # Don't record in localFileIDs because we're not local
                        # No change to the job since there was no caching
                    RealtimeLogger.info('Checking for %d bytes cached', cached)
                    cls._requirementsConcur(job, jobDisk, cached)
                else:
                    if len(writtenFiles) == 0:
                        continue
                    else:
                        fsID, rdelFileSize = random.choice(list(writtenFiles.items()))
                        rdelRandVal = random.random()
                        fileWasCached = job.fileStore.fileIsCached(fsID)
                    if randVal < 0.66:  # Read
                        RealtimeLogger.info('Reading a file with size %d and previous cache status %s: %s', rdelFileSize, str(fileWasCached), fsID)
                        if rdelRandVal <= 0.5:  # Read as mutable, uncached
                            RealtimeLogger.info('Reading as mutable and uncached; should still have %d bytes cached', cached)
                            job.fileStore.readGlobalFile(fsID, '/'.join([work_dir, str(uuid4())]),
                                                         mutable=True, cache=False)
                            localFileIDs[fsID].append('mutable')
                            # No change because the file wasn't cached
                        else:  # Read as immutable
                            RealtimeLogger.info('Reading as immutable and cacheable')
                            job.fileStore.readGlobalFile(fsID, '/'.join([work_dir, str(uuid4())]),
                                                         mutable=False, cache=True)
                            localFileIDs[fsID].append('immutable')
                            jobDisk -= rdelFileSize * 1024 * 1024
                        if not fileWasCached:
                            if job.fileStore.fileIsCached(fsID):
                                RealtimeLogger.info('File was not cached before and is now. Should have %d bytes cached', cached)
                                cached += rdelFileSize * 1024 * 1024
                            else:
                                RealtimeLogger.info('File was not cached before and still is not now. '
                                                    'Should still have %d bytes cached', cached)
                        else:
                            RealtimeLogger.info('File was cached before. Should still have %d bytes cached', cached)
                        cls._requirementsConcur(job, jobDisk, cached)
                    else:  # Delete
                        if rdelRandVal <= 0.5:  # Local Delete
                            if fsID not in list(localFileIDs.keys()):
                                continue
                            RealtimeLogger.info('Deleting a file locally with history %s: %s', localFileIDs[fsID], fsID)
                            job.fileStore.deleteLocalFile(fsID)
                        else:  # Global Delete
                            RealtimeLogger.info('Deleting a file globally: %s', fsID)
                            job.fileStore.deleteGlobalFile(fsID)
                            try:
                                job.fileStore.readGlobalFile(fsID)
                            except FileNotFoundError as err:
                                pass
                            except:
                                raise RuntimeError('Got wrong error type for read of deleted file')
                            else:
                                raise RuntimeError('Able to read deleted file')
                            writtenFiles.pop(fsID)
                        if fsID in list(localFileIDs.keys()):
                            for lFID in localFileIDs[fsID]:
                                if lFID != 'mutable':
                                    jobDisk += rdelFileSize * 1024 * 1024
                            localFileIDs.pop(fsID)
                        if fileWasCached:
                            if not job.fileStore.fileIsCached(fsID):
                                cached -= rdelFileSize * 1024 * 1024
                                RealtimeLogger.info('File was cached before and is not now. Should have %d bytes cached', cached)
                            else:
                                RealtimeLogger.info('File was cached before and still is cached now. '
                                                    'Should still have %d bytes cached', cached)
                        else:
                            RealtimeLogger.info('File was not cached before deletion. Should still have %d bytes cached', cached)
                        cls._requirementsConcur(job, jobDisk, cached)
                i += 1
            return jobDisk, cached

        @staticmethod
        def _requirementsConcur(job, jobDisk, cached):
            """
            Assert the values for job disk and total cached file sizes tracked
            by the file store are equal to the values we expect.
            """

            used = job.fileStore.getCacheUsed()

            if not job.fileStore.cachingIsFree():
                RealtimeLogger.info('Caching is not free; %d bytes are used and %d bytes are expected', used, cached)
                assert used == cached, 'Cache should have %d bytes used, but actually has %d bytes used' % (cached, used)
            else:
                RealtimeLogger.info('Caching is free; %d bytes are used and %d bytes would be expected if caching were not free', used, cached)
                assert used == 0, 'Cache should have nothing in it, but actually has %d bytes used' % used

            jobUnused = job.fileStore.getCacheUnusedJobRequirement()

            assert jobUnused == jobDisk, 'Job should have %d bytes of disk for non-FileStore use but the FileStore reports %d' % (jobDisk, jobUnused)

        # Testing the resumability of a failed worker
        @slow
        def testControlledFailedWorkerRetry(self):
            """
            Conduct a couple of job store operations.  Then die.  Ensure that the restarted job is
            tracking values in the cache state file appropriately.
            """
            workdir = self._createTempDir(purpose='nonLocalDir')
            self.options.retryCount = 1
            jobDiskBytes = 2 * 1024 * 1024 * 1024
            F = Job.wrapJobFn(self._controlledFailTestFn, jobDisk=jobDiskBytes,
                              testDir=workdir,
                              disk=jobDiskBytes)
            G = Job.wrapJobFn(self._probeJobReqs, sigmaJob=100, disk='100Mi')
            F.addChild(G)
            Job.Runner.startToil(F, self.options)

        @staticmethod
        def _controlledFailTestFn(job, jobDisk, testDir):
            """
            This is the aux function for the controlled failed worker test.  It does a couple of
            cache operations, fails, then checks whether the new worker starts with the expected
            value, and whether it computes cache statistics correctly.

            :param float jobDisk: Disk space supplied for this job
            :param str testDir: Testing directory
            """

            # Make sure we actually have the disk size we are supposed to
            job.fileStore.logToMaster('Job is running with %d bytes of disk, %d requested' % (job.disk, jobDisk))
            assert job.disk == jobDisk, 'Job was scheduled with %d bytes but requested %d' % (job.disk, jobDisk)

            cls = hidden.AbstractCachingFileStoreTest
            if os.path.exists(os.path.join(testDir, 'testfile.test')):
                with open(os.path.join(testDir, 'testfile.test'), 'rb') as fH:
                    cached = unpack('d', fH.read())[0]
                RealtimeLogger.info('Loaded expected cache size of %d from testfile.test', cached)
                cls._requirementsConcur(job, jobDisk, cached)
                cls._returnFileTestFn(job, jobDisk, cached, testDir, 20)
            else:
                RealtimeLogger.info('Expecting cache size of 0 because testfile.test is absent')
                modifiedJobReqs, cached = cls._returnFileTestFn(job, jobDisk, 0, testDir, 20)
                with open(os.path.join(testDir, 'testfile.test'), 'wb') as fH:
                    fH.write(pack('d', cached))
                    RealtimeLogger.info('Wrote cache size of %d to testfile.test', cached)
                os.kill(os.getpid(), signal.SIGKILL)

        @slow
        def testRemoveLocalMutablyReadFile(self):
            """
            If a mutably read file is deleted by the user, it is ok.
            """
            self._deleteLocallyReadFilesFn(readAsMutable=True)

        @slow
        def testRemoveLocalImmutablyReadFile(self):
            """
            If an immutably read file is deleted by the user, it is not ok.
            """
            self._deleteLocallyReadFilesFn(readAsMutable=False)

        def _deleteLocallyReadFilesFn(self, readAsMutable):
            self.options.retryCount = 0
            A = Job.wrapJobFn(self._writeFileToJobStoreWithAsserts, isLocalFile=True, memory='10M')
            B = Job.wrapJobFn(self._removeReadFileFn, A.rv(), readAsMutable=readAsMutable,
                              memory='20M')
            A.addChild(B)
            Job.Runner.startToil(A, self.options)

        @staticmethod
        def _removeReadFileFn(job, fileToDelete, readAsMutable):
            """
            Accept a file. Run os.remove on it. Then attempt to delete it locally. This will raise
            an error for files read immutably.
            Then write a new file to the jobstore and try to do the same. This should always raise
            an error

            :param fileToDelete: File written to the job store that is tracked by the cache
            """
            work_dir = job.fileStore.getLocalTempDir()
            # Are we processing the read file or the written file?
            processsingReadFile = True
            # Read in the file
            outfile = job.fileStore.readGlobalFile(fileToDelete, os.path.join(work_dir, 'temp'),
                                                   mutable=readAsMutable)
            tempfile = os.path.join(work_dir, 'tmp.tmp')
            # The first time we run this loop, processsingReadFile is True and fileToDelete is the
            # file read from the job store.  The second time, processsingReadFile is False and
            # fileToDelete is one that was just written in to the job store. Ensure the correct
            # behaviour is seen in both conditions.
            while True:
                os.rename(outfile, tempfile)
                try:
                    job.fileStore.deleteLocalFile(fileToDelete)
                except IllegalDeletionCacheError:
                    job.fileStore.logToMaster('Detected a deleted file %s.' % fileToDelete)
                    os.rename(tempfile, outfile)
                else:
                    # If we are processing the write test, or if we are testing the immutably read
                    # file, we should not reach here.
                    assert processsingReadFile and readAsMutable
                if processsingReadFile:
                    processsingReadFile = False
                    # Write a file
                    with open(os.path.join(work_dir, str(uuid4())), 'wb') as testFile:
                        testFile.write(os.urandom(1 * 1024 * 1024))
                    fileToDelete = job.fileStore.writeGlobalFile(testFile.name)
                    outfile = testFile.name
                else:
                    break

        def testDeleteLocalFile(self):
            """
            Test the deletion capabilities of deleteLocalFile
            """
            self.options.retryCount = 0
            workdir = self._createTempDir(purpose='nonLocalDir')
            A = Job.wrapJobFn(self._deleteLocalFileFn, nonLocalDir=workdir)
            Job.Runner.startToil(A, self.options)

        @staticmethod
        def _deleteLocalFileFn(job, nonLocalDir):
            """
            Test deleteLocalFile on a local write, non-local write, read, mutable read, and bogus
            jobstore IDs.
            """
            work_dir = job.fileStore.getLocalTempDir()
            # Write local file
            with open(os.path.join(work_dir, str(uuid4())), 'wb') as localFile:
                localFile.write(os.urandom(1 * 1024 * 1024))
            localFsID = job.fileStore.writeGlobalFile(localFile.name)
            # write Non-Local File
            with open(os.path.join(nonLocalDir, str(uuid4())), 'wb') as nonLocalFile:
                nonLocalFile.write(os.urandom(1 * 1024 * 1024))
            nonLocalFsID = job.fileStore.writeGlobalFile(nonLocalFile.name)
            # Delete fsid of local file. The file should be deleted
            job.fileStore.deleteLocalFile(localFsID)
            assert not os.path.exists(localFile.name)
            # Delete fsid of non-local file. It should fail and the file should persist
            try:
                job.fileStore.deleteLocalFile(nonLocalFsID)
            except OSError as e:
                if e.errno != errno.ENOENT:
                    # This is supposed to produce an
                    # ENOENT. If it doesn't something is
                    # broken.
                    raise
            else:
                assert False, "Error should have been raised"
            assert os.path.exists(nonLocalFile.name)
            # Read back one file and then delete it
            readBackFile1 = job.fileStore.readGlobalFile(localFsID)
            job.fileStore.deleteLocalFile(localFsID)
            assert not os.path.exists(readBackFile1)
            # Read back one file with 2 different names and then delete it. Assert both get deleted
            readBackFile1 = job.fileStore.readGlobalFile(localFsID)
            readBackFile2 = job.fileStore.readGlobalFile(localFsID)
            job.fileStore.deleteLocalFile(localFsID)
            assert not os.path.exists(readBackFile1)
            assert not os.path.exists(readBackFile2)
            # Try to get a non-FileID that doesn't exist.
            try:
                job.fileStore.readGlobalFile('bogus')
            except NoSuchFileException:
                # TODO: We would like to require TypeError, but for Cactus
                # support we have to accept non-FileIDs.
                pass
            else:
                raise RuntimeError("Managed to get a file from a non-FileID")
            # Try to get a FileID for something that doesn't exist
            try:
                job.fileStore.readGlobalFile(FileID('bogus', 4096))
            except NoSuchFileException:
                pass
            else:
                raise RuntimeError("Managed to read a non-existent file")

        def testSimultaneousReadsUncachedStream(self):
            """
            Test many simultaneous read attempts on a file created via a stream
            directly to the job store.
            """
            self.options.retryCount = 0
            self.options.disableChaining = True

            # Make a file
            parent = Job.wrapJobFn(self._createUncachedFileStream)
            # Now make a bunch of children fight over it
            for i in range(30):
                parent.addChildJobFn(self._readFileWithDelay, parent.rv())

            Job.Runner.startToil(parent, self.options)

        @staticmethod
        def _createUncachedFileStream(job):
            """
            Create and return a FileID for a non-cached file written via a stream.
            """

            messageBytes = b'This is a test file\n'

            with job.fileStore.jobStore.write_file_stream() as (out, idString):
                # Write directly to the job store so the caching file store doesn't even see it.
                # TODO: If we ever change how the caching file store does its IDs we will have to change this.
                out.write(messageBytes)

            # Now make a file ID
            fileID = FileID(idString, len(messageBytes))

            return fileID

        @staticmethod
        def _readFileWithDelay(job, fileID, cores=0.1, memory=50 * 1024 * 1024, disk=50 * 1024 * 1024):
            """
            Read a file from the CachingFileStore with a delay imposed on the download.
            Should create contention.

            Has low requirements so we can run a lot of copies at once.
            """

            # Make sure the file store delays
            # Delay needs to be longer than the timeout for sqlite locking in the file store.
            job.fileStore.forceDownloadDelay = 120

            readStart = datetime.datetime.now()
            logger.debug('Begin read at %s', str(readStart))

            localPath = job.fileStore.readGlobalFile(fileID, cache=True, mutable=True)

            readEnd = datetime.datetime.now()
            logger.debug('End read at %s: took %f seconds', str(readEnd), (readEnd - readStart).total_seconds())

            with open(localPath, 'rb') as fh:
                text = fh.read().decode('utf-8').strip()
            logger.debug('Got file contents: %s', text)



class NonCachingFileStoreTestWithFileJobStore(hidden.AbstractNonCachingFileStoreTest):
    jobStoreType = 'file'

@pytest.mark.timeout(1000)
class CachingFileStoreTestWithFileJobStore(hidden.AbstractCachingFileStoreTest):
    jobStoreType = 'file'


@needs_aws_ec2
class NonCachingFileStoreTestWithAwsJobStore(hidden.AbstractNonCachingFileStoreTest):
    jobStoreType = 'aws'


@slow
@needs_aws_ec2
@pytest.mark.timeout(1000)
class CachingFileStoreTestWithAwsJobStore(hidden.AbstractCachingFileStoreTest):
    jobStoreType = 'aws'


@needs_google
class NonCachingFileStoreTestWithGoogleJobStore(hidden.AbstractNonCachingFileStoreTest):
    jobStoreType = 'google'


@slow
@needs_google
@pytest.mark.timeout(1000)
class CachingFileStoreTestWithGoogleJobStore(hidden.AbstractCachingFileStoreTest):
    jobStoreType = 'google'


def _exportStaticMethodAsGlobalFunctions(cls):
    """
    Define utility functions because Toil can't pickle static methods. Note that this relies on
    the convention that the first argument of a job function is named 'job'.
    """
    for name, kind, clazz, value in inspect.classify_class_attrs(cls):
        if kind == 'static method' and name != '__new__':  # __new__ became static in 3.7
            method = value.__func__
            args = inspect.getfullargspec(method).args
            if args and args[0] == 'job':
                globals()[name] = method


_exportStaticMethodAsGlobalFunctions(hidden.AbstractFileStoreTest)
_exportStaticMethodAsGlobalFunctions(hidden.AbstractCachingFileStoreTest)
_exportStaticMethodAsGlobalFunctions(hidden.AbstractNonCachingFileStoreTest)



import logging
import multiprocessing
import os
import random
import time
import traceback
from functools import partial

from toil.lib.threading import (LastProcessStandingArena,
                                cpu_count,
                                global_mutex)
from toil.test import ToilTest

log = logging.getLogger(__name__)

class ThreadingTest(ToilTest):
    """Test Toil threading/synchronization tools."""
    def testGlobalMutexOrdering(self):
        for it in range(10):
            log.info('Iteration %d', it)

            scope = self._createTempDir()
            mutex = 'mutex'
            # Use processes (as opposed to threads) to prevent GIL from ordering things artificially
            pool = multiprocessing.Pool(processes=cpu_count())
            try:
                numTasks = 100
                results = pool.map_async(
                    func=partial(_testGlobalMutexOrderingTask, scope, mutex),
                    iterable=list(range(numTasks)))
                results = results.get()
            finally:
                pool.close()
                pool.join()

            self.assertEqual(len(results), numTasks)
            for item in results:
                # Make sure all workers say they succeeded
                self.assertEqual(item, True)

    def testLastProcessStanding(self):
        for it in range(10):
            log.info('Iteration %d', it)

            scope = self._createTempDir()
            arena_name = 'thunderdome'
            # Use processes (as opposed to threads) to prevent GIL from ordering things artificially
            pool = multiprocessing.Pool(processes=cpu_count())
            try:
                numTasks = 100
                results = pool.map_async(
                    func=partial(_testLastProcessStandingTask, scope, arena_name),
                    iterable=list(range(numTasks)))
                results = results.get()
            finally:
                pool.close()
                pool.join()

            self.assertEqual(len(results), numTasks)
            for item in results:
                # Make sure all workers say they succeeded
                self.assertEqual(item, True)
            for filename in os.listdir(scope):
                assert not filename.startswith('precious'), f"File {filename} still exists"

def _testGlobalMutexOrderingTask(scope, mutex, number):
    try:
        # We will all fight over the potato
        potato = os.path.join(scope, 'potato')

        with global_mutex(scope, mutex):
            log.info('PID %d = num %d running', os.getpid(), number)
            assert not os.path.exists(potato), "We see someone else holding the potato file"

            # Put our name there
            with open(potato, 'w') as out_stream:
                out_stream.write(str(number))

            # Wait
            time.sleep(random.random() * 0.01)

            # Make sure our name is still there
            with open(potato) as in_stream:
                seen = in_stream.read().rstrip()
                assert seen == str(number), f"We are {number} but {seen} stole our potato!"

            os.unlink(potato)
            assert not os.path.exists(potato), "We left the potato behind"
            log.info('PID %d = num %d dropped potato', os.getpid(), number)
        return True
    except:
        traceback.print_exc()
        return False

def _testLastProcessStandingTask(scope, arena_name, number):
    try:
        arena = LastProcessStandingArena(scope, arena_name)

        arena.enter()
        log.info('PID %d = num %d entered arena', os.getpid(), number)
        try:
            # We all make files
            my_precious = os.path.join(scope, 'precious' + str(number))

            # Put our name there
            with open(my_precious, 'w') as out_stream:
                out_stream.write(str(number))

            # Wait
            time.sleep(random.random() * 0.01)

            # Make sure our file is still there unmodified
            assert os.path.exists(my_precious), f"Precious file {my_precious} has been stolen!"
            with open(my_precious) as in_stream:
                seen = in_stream.read().rstrip()
                assert seen == str(number), f"We are {number} but saw {seen} in our precious file!"
        finally:
            was_last = False
            for _ in arena.leave():
                was_last = True
                log.info('PID %d = num %d is last standing', os.getpid(), number)

                # Clean up all the files
                for filename in os.listdir(scope):
                    if filename.startswith('precious'):
                        log.info('PID %d = num %d cleaning up %s', os.getpid(), number, filename)
                        os.unlink(os.path.join(scope, filename))

            log.info('PID %d = num %d left arena', os.getpid(), number)

        return True
    except:
        traceback.print_exc()
        return False
